"""
Node module
"""

from enum import Enum
from typing import Optional, Union, TYPE_CHECKING

from slither.all_exceptions import SlitherException
from slither.core.declarations import Contract, Function, FunctionContract
from slither.core.declarations.solidity_variables import (
    SolidityVariable,
    SolidityFunction,
)
from slither.core.expressions.expression import Expression
from slither.core.solidity_types import ElementaryType
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.variable import Variable
from slither.slithir.convert import convert_expression
from slither.slithir.operations import (
    HighLevelCall,
    Index,
    InternalCall,
    Length,
    LibraryCall,
    LowLevelCall,
    Member,
    OperationWithLValue,
    Phi,
    PhiCallback,
    SolidityCall,
    Return,
    Operation,
)
from slither.slithir.utils.utils import RVALUE
from slither.slithir.variables import (
    Constant,
    LocalIRVariable,
    ReferenceVariable,
    StateIRVariable,
    TemporaryVariable,
    TupleVariable,
)

if TYPE_CHECKING:
    from slither.slithir.variables.variable import SlithIRVariable
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.core.cfg.scope import Scope
    from slither.core.scope.scope import FileScope


###################################################################################
###################################################################################
# region NodeType
###################################################################################
###################################################################################


class NodeType(Enum):
    ENTRYPOINT = "ENTRY_POINT"  # no expression

    # Nodes that may have an expression

    EXPRESSION = "EXPRESSION"  # normal case
    RETURN = "RETURN"  # RETURN may contain an expression
    IF = "IF"
    VARIABLE = "NEW VARIABLE"  # Variable declaration
    ASSEMBLY = "INLINE ASM"
    ENDASSEMBLY = "END INLINE ASM"
    IFLOOP = "IF_LOOP"

    # Nodes where control flow merges
    # Can have phi IR operation
    ENDIF = "END_IF"  # ENDIF node source mapping points to the if/else "body"
    STARTLOOP = "BEGIN_LOOP"  # STARTLOOP node source mapping points to the entire loop "body"
    ENDLOOP = "END_LOOP"  # ENDLOOP node source mapping points to the entire loop "body"

    # Below the nodes do not have an expression but are used to expression CFG structure.

    # Absorbing node
    THROW = "THROW"

    # Loop related nodes
    BREAK = "BREAK"
    CONTINUE = "CONTINUE"

    # Only modifier node
    PLACEHOLDER = "_"

    TRY = "TRY"
    CATCH = "CATCH"

    # Node not related to the CFG
    # Use for state variable declaration
    OTHER_ENTRYPOINT = "OTHER_ENTRYPOINT"


# endregion


class Node(SourceMapping):
    """
    Node class

    """

    def __init__(
        self,
        node_type: NodeType,
        node_id: int,
        scope: Union["Scope", "Function"],
        file_scope: "FileScope",
    ) -> None:
        super().__init__()
        self._node_type = node_type

        # TODO: rename to explicit CFG
        self._sons: list[Node] = []
        self._fathers: list[Node] = []

        ## Dominators info
        # Dominators nodes
        self._dominators: set[Node] = set()
        self._immediate_dominator: Node | None = None
        ## Nodes of the dominators tree
        # self._dom_predecessors = set()
        self._dom_successors: set[Node] = set()
        # Dominance frontier
        self._dominance_frontier: set[Node] = set()
        # Phi origin
        # key are variable name
        self._phi_origins_state_variables: dict[str, tuple[StateVariable, set[Node]]] = {}
        self._phi_origins_local_variables: dict[str, tuple[LocalVariable, set[Node]]] = {}
        # self._phi_origins_member_variables: Dict[str, Tuple[MemberVariable, Set["Node"]]] = {}

        self._expression: Expression | None = None
        self._variable_declaration: LocalVariable | None = None
        self._node_id: int = node_id

        self._vars_written: list[Variable] = []
        self._vars_read: list[Variable | SolidityVariable] = []

        self._ssa_vars_written: list[SlithIRVariable] = []
        self._ssa_vars_read: list[SlithIRVariable] = []

        self._internal_calls: list[InternalCall] = []  # contains solidity calls
        self._solidity_calls: list[SolidityCall] = []
        self._high_level_calls: list[tuple[Contract, HighLevelCall]] = []  # contains library calls
        self._library_calls: list[LibraryCall] = []
        self._low_level_calls: list[LowLevelCall] = []
        self._external_calls_as_expressions: list[Expression] = []
        self._internal_calls_as_expressions: list[Expression] = []
        self._irs: list[Operation] = []
        self._all_slithir_operations: list[Operation] | None = None
        self._irs_ssa: list[Operation] = []

        self._state_vars_written: list[StateVariable] = []
        self._state_vars_read: list[StateVariable] = []
        self._solidity_vars_read: list[SolidityVariable] = []

        self._ssa_state_vars_written: list[StateIRVariable] = []
        self._ssa_state_vars_read: list[StateIRVariable] = []

        self._local_vars_read: list[LocalVariable] = []
        self._local_vars_written: list[LocalVariable] = []

        self._slithir_vars: set[
            SlithIRVariable | ReferenceVariable | TemporaryVariable | TupleVariable
        ] = set()  # non SSA

        self._ssa_local_vars_read: list[LocalIRVariable] = []
        self._ssa_local_vars_written: list[LocalIRVariable] = []

        self._expression_vars_written: list[Expression] = []
        self._expression_vars_read: list[Expression] = []
        self._expression_calls: list[Expression] = []

        # Computed on the fly, can be True of False
        self._can_reenter: bool | None = None
        self._can_send_eth: bool | None = None

        self._asm_source_code: str | dict | None = None

        self.scope: Scope | Function = scope
        self.file_scope: FileScope = file_scope
        self._function: Function | None = None

        self._is_reachable: bool = False

    ###################################################################################
    ###################################################################################
    # region General's properties
    ###################################################################################
    ###################################################################################

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self.function.compilation_unit

    @property
    def node_id(self) -> int:
        """Unique node id."""
        return self._node_id

    @property
    def type(self) -> NodeType:
        """
        NodeType: type of the node
        """
        return self._node_type

    @type.setter
    def type(self, new_type: NodeType) -> None:
        self._node_type = new_type

    @property
    def will_return(self) -> bool:
        if not self.sons and self.type != NodeType.THROW:
            solidity_calls = [ir.function for ir in self.solidity_calls]
            if SolidityFunction("revert()") not in solidity_calls:
                if SolidityFunction("revert(string)") not in solidity_calls:
                    return True
        return False

    def set_function(self, function: "Function") -> None:
        self._function = function

    @property
    def function(self) -> "Function":
        return self._function

    @property
    def is_reachable(self) -> bool:
        return self._is_reachable

    def set_is_reachable(self, new_is_reachable: bool) -> None:
        self._is_reachable = new_is_reachable

    # endregion
    ###################################################################################
    ###################################################################################
    # region Variables
    ###################################################################################
    ###################################################################################

    @property
    def variables_read(self) -> list[Variable | SolidityVariable]:
        """
        list(Variable): Variables read (local/state/solidity)
        """
        return list(self._vars_read)

    @property
    def state_variables_read(self) -> list[StateVariable]:
        """
        list(StateVariable): State variables read
        """
        return list(self._state_vars_read)

    @property
    def local_variables_read(self) -> list[LocalVariable]:
        """
        list(LocalVariable): Local variables read
        """
        return list(self._local_vars_read)

    @property
    def solidity_variables_read(self) -> list[SolidityVariable]:
        """
        list(SolidityVariable): State variables read
        """
        return list(self._solidity_vars_read)

    @property
    def ssa_variables_read(self) -> list["SlithIRVariable"]:
        """
        list(Variable): Variables read (local/state/solidity)
        """
        return list(self._ssa_vars_read)

    @property
    def ssa_state_variables_read(self) -> list[StateIRVariable]:
        """
        list(StateVariable): State variables read
        """
        return list(self._ssa_state_vars_read)

    @property
    def ssa_local_variables_read(self) -> list[LocalIRVariable]:
        """
        list(LocalVariable): Local variables read
        """
        return list(self._ssa_local_vars_read)

    @property
    def variables_read_as_expression(self) -> list[Expression]:
        return self._expression_vars_read

    @variables_read_as_expression.setter
    def variables_read_as_expression(self, exprs: list[Expression]) -> None:
        self._expression_vars_read = exprs

    @property
    def slithir_variables(
        self,
    ) -> list[Union["SlithIRVariable", ReferenceVariable, TemporaryVariable, TupleVariable]]:
        return list(self._slithir_vars)

    @property
    def variables_written(self) -> list[Variable]:
        """
        list(Variable): Variables written (local/state/solidity)
        """
        return list(self._vars_written)

    @property
    def state_variables_written(self) -> list[StateVariable]:
        """
        list(StateVariable): State variables written
        """
        return list(self._state_vars_written)

    @property
    def local_variables_written(self) -> list[LocalVariable]:
        """
        list(LocalVariable): Local variables written
        """
        return list(self._local_vars_written)

    @property
    def ssa_variables_written(self) -> list["SlithIRVariable"]:
        """
        list(Variable): Variables written (local/state/solidity)
        """
        return list(self._ssa_vars_written)

    @property
    def ssa_state_variables_written(self) -> list[StateIRVariable]:
        """
        list(StateVariable): State variables written
        """
        return list(self._ssa_state_vars_written)

    @property
    def ssa_local_variables_written(self) -> list[LocalIRVariable]:
        """
        list(LocalVariable): Local variables written
        """
        return list(self._ssa_local_vars_written)

    @property
    def variables_written_as_expression(self) -> list[Expression]:
        return self._expression_vars_written

    @variables_written_as_expression.setter
    def variables_written_as_expression(self, exprs: list[Expression]) -> None:
        self._expression_vars_written = exprs

    # endregion
    ###################################################################################
    ###################################################################################
    # region Calls
    ###################################################################################
    ###################################################################################

    @property
    def internal_calls(self) -> list[InternalCall]:
        """
        list(InternalCall): List of IR operations with internal/solidity function calls
        """
        return list(self._internal_calls)

    @property
    def solidity_calls(self) -> list[SolidityCall]:
        """
        list(SolidityCall): List of IR operations with solidity calls
        """
        return list(self._solidity_calls)

    @property
    def high_level_calls(self) -> list[HighLevelCall]:
        """
        list(HighLevelCall): List of IR operations with high level calls (external calls).
        Include library calls
        """
        return list(self._high_level_calls)

    @property
    def library_calls(self) -> list[LibraryCall]:
        """
        list(LibraryCall): List of IR operations with library calls.
        """
        return list(self._library_calls)

    @property
    def low_level_calls(self) -> list[LowLevelCall]:
        """
        list(LowLevelCall): List of IR operations with low_level call
        """
        return list(self._low_level_calls)

    @property
    def external_calls_as_expressions(self) -> list[Expression]:
        """
        list(CallExpression): List of message calls (that creates a transaction)
        """
        return self._external_calls_as_expressions

    @external_calls_as_expressions.setter
    def external_calls_as_expressions(self, exprs: list[Expression]) -> None:
        self._external_calls_as_expressions = exprs

    @property
    def internal_calls_as_expressions(self) -> list[Expression]:
        """
        list(CallExpression): List of internal calls (that dont create a transaction)
        """
        return self._internal_calls_as_expressions

    @internal_calls_as_expressions.setter
    def internal_calls_as_expressions(self, exprs: list[Expression]) -> None:
        self._internal_calls_as_expressions = exprs

    @property
    def calls_as_expression(self) -> list[Expression]:
        return list(self._expression_calls)

    @calls_as_expression.setter
    def calls_as_expression(self, exprs: list[Expression]) -> None:
        self._expression_calls = exprs

    def can_reenter(self, callstack: list[Function | Variable] | None = None) -> bool:
        """
        Check if the node can re-enter
        Do not consider CREATE as potential re-enter, but check if the
        destination's constructor can contain a call (recurs. follow nested CREATE)
        For Solidity > 0.5, filter access to public variables and constant/pure/view
        For call to this. check if the destination can re-enter
        Do not consider Send/Transfer as there is not enough gas
        :param callstack: used internally to check for recursion
        :return bool:
        """

        from slither.slithir.operations import Call

        if self._can_reenter is None:
            self._can_reenter = False
            for ir in self.irs:
                if isinstance(ir, Call) and ir.can_reenter(callstack):
                    self._can_reenter = True
                    return True
        return self._can_reenter

    def can_send_eth(self) -> bool:
        """
        Check if the node can send eth
        :return bool:
        """

        from slither.slithir.operations import Call

        if self._can_send_eth is None:
            self._can_send_eth = False
            for ir in self.all_slithir_operations():
                if isinstance(ir, Call) and ir.can_send_eth():
                    self._can_send_eth = True
                    return True
        return self._can_send_eth

    # endregion
    ###################################################################################
    ###################################################################################
    # region Expressions
    ###################################################################################
    ###################################################################################

    @property
    def expression(self) -> Expression | None:
        """
        Expression: Expression of the node
        """
        return self._expression

    def add_expression(self, expression: Expression, bypass_verif_empty: bool = False) -> None:
        assert self._expression is None or bypass_verif_empty
        self._expression = expression

    def add_variable_declaration(self, var: LocalVariable) -> None:
        assert self._variable_declaration is None
        self._variable_declaration = var
        if var.expression:
            self._vars_written += [var]
            self._local_vars_written += [var]

    @property
    def variable_declaration(self) -> LocalVariable | None:
        """
        Returns:
            LocalVariable
        """
        return self._variable_declaration

    # endregion
    ###################################################################################
    ###################################################################################
    # region Summary information
    ###################################################################################
    ###################################################################################

    def contains_require_or_assert(self) -> bool:
        """
            Check if the node has a require or assert call
        Returns:
            bool: True if the node has a require or assert call
        """
        return any(
            ir.function.name
            in ["require(bool)", "require(bool,string)", "require(bool,error)", "assert(bool)"]
            for ir in self.internal_calls
        )

    def contains_if(self, include_loop: bool = True) -> bool:
        """
            Check if the node is a IF node
        Returns:
            bool: True if the node is a conditional node (IF or IFLOOP)
        """
        if include_loop:
            return self.type in [NodeType.IF, NodeType.IFLOOP]
        return self.type == NodeType.IF

    def is_conditional(self, include_loop: bool = True) -> bool:
        """
            Check if the node is a conditional node
            A conditional node is either a IF or a require/assert or a RETURN bool
        Returns:
            bool: True if the node is a conditional node
        """
        if self.contains_if(include_loop) or self.contains_require_or_assert():
            return True
        if self.irs:
            last_ir = self.irs[-1]
            if last_ir:
                if isinstance(last_ir, Return):
                    for r in last_ir.read:
                        if r.type == ElementaryType("bool"):
                            return True
        return False

    # endregion
    ###################################################################################
    ###################################################################################
    # region EVM
    ###################################################################################
    ###################################################################################

    @property
    def inline_asm(self) -> str | dict | None:
        return self._asm_source_code

    def add_inline_asm(self, asm: str | dict) -> None:
        self._asm_source_code = asm

    # endregion
    ###################################################################################
    ###################################################################################
    # region Graph
    ###################################################################################
    ###################################################################################

    def add_father(self, father: "Node") -> None:
        """Add a father node

        Args:
            father: father to add
        """
        self._fathers.append(father)

    def set_fathers(self, fathers: list["Node"]) -> None:
        """Set the father nodes

        Args:
            fathers: list of fathers to add
        """
        self._fathers = fathers

    @property
    def fathers(self) -> list["Node"]:
        """Returns the father nodes

        Returns:
            list(Node): list of fathers
        """
        return list(self._fathers)

    def remove_father(self, father: "Node") -> None:
        """Remove the father node. Do nothing if the node is not a father

        Args:
            :param father:
        """
        self._fathers = [x for x in self._fathers if x.node_id != father.node_id]

    def remove_son(self, son: "Node") -> None:
        """Remove the son node. Do nothing if the node is not a son

        Args:
            :param son:
        """
        self._sons = [x for x in self._sons if x.node_id != son.node_id]

    def add_son(self, son: "Node") -> None:
        """Add a son node

        Args:
            son: son to add
        """
        self._sons.append(son)

    def replace_son(self, ori_son: "Node", new_son: "Node") -> None:
        """Replace a son node. Do nothing if the node to replace is not a son

        Args:
            ori_son: son to replace
            new_son: son to replace with
        """
        for i, s in enumerate(self._sons):
            if s.node_id == ori_son.node_id:
                idx = i
                break
        else:
            return
        self._sons[idx] = new_son

    def set_sons(self, sons: list["Node"]) -> None:
        """Set the son nodes

        Args:
            sons: list of fathers to add
        """
        self._sons = sons

    @property
    def sons(self) -> list["Node"]:
        """Returns the son nodes

        Returns:
            list(Node): list of sons
        """
        return list(self._sons)

    @property
    def son_true(self) -> Optional["Node"]:
        if self.type in [NodeType.IF, NodeType.IFLOOP]:
            return self._sons[0]
        return None

    @property
    def son_false(self) -> Optional["Node"]:
        if self.type in [NodeType.IF, NodeType.IFLOOP] and len(self._sons) >= 1:
            return self._sons[1]
        return None

    # endregion
    ###################################################################################
    ###################################################################################
    # region SlithIR
    ###################################################################################
    ###################################################################################

    @property
    def irs(self) -> list[Operation]:
        """Returns the slithIR representation

        return
            list(slithIR.Operation)
        """
        return self._irs

    @property
    def irs_ssa(self) -> list[Operation]:
        """Returns the slithIR representation with SSA

        return
            list(slithIR.Operation)
        """
        return self._irs_ssa

    @irs_ssa.setter
    def irs_ssa(self, irs: list[Operation]) -> None:
        self._irs_ssa = irs

    def add_ssa_ir(self, ir: Operation) -> None:
        """
        Use to place phi operation
        """
        ir.set_node(self)  # type: ignore
        self._irs_ssa.append(ir)

    def slithir_generation(self) -> None:
        if self.expression:
            expression = self.expression
            self._irs = convert_expression(expression, self)  # type:ignore

        self._find_read_write_call()

    def all_slithir_operations(self) -> list[Operation]:
        if self._all_slithir_operations is None:
            irs = list(self.irs)
            for ir in self.irs:
                if isinstance(ir, InternalCall):
                    irs += ir.function.all_slithir_operations()
            self._all_slithir_operations = irs
        return self._all_slithir_operations

    @staticmethod
    def _is_non_slithir_var(var: Variable) -> bool:
        return not isinstance(var, (Constant, ReferenceVariable, TemporaryVariable, TupleVariable))

    @staticmethod
    def _is_valid_slithir_var(var: Variable) -> bool:
        return isinstance(var, (ReferenceVariable, TemporaryVariable, TupleVariable))

    # endregion
    ###################################################################################
    ###################################################################################
    # region Dominators
    ###################################################################################
    ###################################################################################

    @property
    def dominators(self) -> set["Node"]:
        """
        Returns:
            set(Node)
        """
        return self._dominators

    @dominators.setter
    def dominators(self, dom: set["Node"]) -> None:
        self._dominators = dom

    @property
    def immediate_dominator(self) -> Optional["Node"]:
        """
        Returns:
            Node or None
        """
        return self._immediate_dominator

    @immediate_dominator.setter
    def immediate_dominator(self, idom: "Node") -> None:
        self._immediate_dominator = idom

    @property
    def dominance_frontier(self) -> set["Node"]:
        """
        Returns:
            set(Node)
        """
        return self._dominance_frontier

    @dominance_frontier.setter
    def dominance_frontier(self, doms: set["Node"]) -> None:
        """
        Returns:
            set(Node)
        """
        self._dominance_frontier = doms

    @property
    def dominator_successors(self) -> set["Node"]:
        return self._dom_successors

    @property
    def dominance_exploration_ordered(self) -> list["Node"]:
        """
        Sorted list of all the nodes to explore to follow the dom
        :return: list(nodes)
        """
        # Explore direct dominance
        to_explore = sorted(self.dominator_successors, key=lambda x: x.node_id)

        # Explore dominance frontier
        # The frontier is the limit where this node dominates
        # We need to explore it because the sub of the direct dominance
        # Might not be dominator of their own sub
        to_explore += sorted(self.dominance_frontier, key=lambda x: x.node_id)
        return to_explore

    # endregion
    ###################################################################################
    ###################################################################################
    # region Phi operation
    ###################################################################################
    ###################################################################################

    @property
    def phi_origins_local_variables(
        self,
    ) -> dict[str, tuple[LocalVariable, set["Node"]]]:
        return self._phi_origins_local_variables

    @property
    def phi_origins_state_variables(
        self,
    ) -> dict[str, tuple[StateVariable, set["Node"]]]:
        return self._phi_origins_state_variables

    # @property
    # def phi_origin_member_variables(self) -> Dict[str, Tuple[MemberVariable, Set["Node"]]]:
    #     return self._phi_origins_member_variables

    def add_phi_origin_local_variable(self, variable: LocalVariable, node: "Node") -> None:
        if variable.name not in self._phi_origins_local_variables:
            assert variable.name
            self._phi_origins_local_variables[variable.name] = (variable, set())
        (v, nodes) = self._phi_origins_local_variables[variable.name]
        assert v == variable
        nodes.add(node)

    def add_phi_origin_state_variable(self, variable: StateVariable, node: "Node") -> None:
        if variable.canonical_name not in self._phi_origins_state_variables:
            self._phi_origins_state_variables[variable.canonical_name] = (
                variable,
                set(),
            )
        (v, nodes) = self._phi_origins_state_variables[variable.canonical_name]
        assert v == variable
        nodes.add(node)

    # def add_phi_origin_member_variable(self, variable: MemberVariable, node: "Node"):
    #     if variable.name not in self._phi_origins_member_variables:
    #         self._phi_origins_member_variables[variable.name] = (variable, set())
    #     (v, nodes) = self._phi_origins_member_variables[variable.name]
    #     assert v == variable
    #     nodes.add(node)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Analyses
    ###################################################################################
    ###################################################################################

    def _find_read_write_call(self) -> None:
        for ir in self.irs:
            self._slithir_vars |= {v for v in ir.read if self._is_valid_slithir_var(v)}

            if isinstance(ir, OperationWithLValue):
                var = ir.lvalue
                if var and self._is_valid_slithir_var(var):
                    # The type is checked by is_valid_slithir_var
                    self._slithir_vars.add(var)  # type: ignore

            if not isinstance(ir, (Phi, Index, Member)):
                self._vars_read += [v for v in ir.read if self._is_non_slithir_var(v)]
                for var in ir.read:
                    if isinstance(var, ReferenceVariable):
                        self._vars_read.append(var.points_to_origin)
            elif isinstance(ir, (Member, Index)):
                # TODO investigate types for member variable left
                var = ir.variable_left if isinstance(ir, Member) else ir.variable_right
                if var and self._is_non_slithir_var(var):
                    self._vars_read.append(var)
                if isinstance(var, ReferenceVariable):
                    origin = var.points_to_origin
                    if self._is_non_slithir_var(origin):
                        self._vars_read.append(origin)

            if isinstance(ir, OperationWithLValue):
                if isinstance(ir, (Index, Member, Length)):
                    continue  # Don't consider Member and Index operations -> ReferenceVariable
                var = ir.lvalue
                if isinstance(var, ReferenceVariable):
                    var = var.points_to_origin
                if var and self._is_non_slithir_var(var):
                    self._vars_written.append(var)

            if isinstance(ir, InternalCall):
                self._internal_calls.append(ir)
            if isinstance(ir, SolidityCall):
                # TODO: consider removing dependency of solidity_call to internal_call
                self._solidity_calls.append(ir)
                self._internal_calls.append(ir)
            if isinstance(ir, LowLevelCall):
                assert isinstance(ir.destination, (Variable, SolidityVariable))
                self._low_level_calls.append(ir)
            elif isinstance(ir, HighLevelCall) and not isinstance(ir, LibraryCall):
                # Todo investigate this if condition
                # It does seem right to compare against a contract
                # This might need a refactoring
                if isinstance(ir.destination.type, Contract):
                    self._high_level_calls.append((ir.destination.type, ir))
                elif ir.destination == SolidityVariable("this"):
                    func = self.function
                    # Can't use this in a top level function
                    assert isinstance(func, FunctionContract)
                    self._high_level_calls.append((func.contract, ir))
                else:
                    try:
                        # Todo this part needs more tests and documentation
                        self._high_level_calls.append((ir.destination.type.type, ir))
                    except AttributeError as error:
                        raise SlitherException(
                            f"Function not found on IR: {ir}.\nNode: {self} ({self.source_mapping})\nFunction: {self.function}\nPlease try compiling with a recent Solidity version. {error}"
                        )
            elif isinstance(ir, LibraryCall):
                assert isinstance(ir.destination, Contract)
                assert isinstance(ir.function, Function)
                self._high_level_calls.append((ir.destination, ir))
                self._library_calls.append(ir)

        self._vars_read = list(set(self._vars_read))
        self._state_vars_read = [v for v in self._vars_read if isinstance(v, StateVariable)]
        self._local_vars_read = [v for v in self._vars_read if isinstance(v, LocalVariable)]
        self._solidity_vars_read = [
            v_ for v_ in self._vars_read if isinstance(v_, SolidityVariable)
        ]
        self._vars_written = list(set(self._vars_written))
        self._state_vars_written = [v for v in self._vars_written if isinstance(v, StateVariable)]
        self._local_vars_written = [v for v in self._vars_written if isinstance(v, LocalVariable)]
        self._internal_calls = list(set(self._internal_calls))
        self._solidity_calls = list(set(self._solidity_calls))
        self._high_level_calls = list(set(self._high_level_calls))
        self._library_calls = list(set(self._library_calls))
        self._low_level_calls = list(set(self._low_level_calls))

    @staticmethod
    def _convert_ssa(v: Variable) -> StateVariable | LocalVariable | None:
        non_ssa_var: StateVariable | LocalVariable | None
        if isinstance(v, StateIRVariable):
            contract = v.contract
            assert v.name
            non_ssa_var = contract.get_state_variable_from_name(v.name)
            return non_ssa_var
        assert isinstance(v, LocalIRVariable)
        function = v.function
        assert v.name
        non_ssa_var = function.get_local_variable_from_name(v.name)
        return non_ssa_var

    def _update_read_using_ssa(self, ir: Operation) -> None:
        """
        Update self._ssa_vars_read
        This look for all operations that read a IRvariable
        It uses the result of the storage pointer
        - For "normal" operation, the read are mostly everything in ir.read
        - For "index", the read is the left part (the right part being a reference variable)
        - For Phi, nothing is considered read

        """

        # For variable read, phi and index have special treatments
        # Phi don't lead to values read
        # Index leads to read the variable right (the left variable is a ref variable, not the actual object)
        # Not that Member is a normal operation here, given we filter out constant by checking for the IRvaraible
        if not isinstance(ir, (Phi, Index)):
            self._ssa_vars_read += [
                v for v in ir.read if isinstance(v, (StateIRVariable, LocalIRVariable))
            ]
            for var in ir.read:
                if isinstance(var, ReferenceVariable):
                    origin = var.points_to_origin
                    if isinstance(origin, (StateIRVariable, LocalIRVariable)):
                        self._ssa_vars_read.append(origin)

                # If we read from a storage variable (outside of phi operator)
                if isinstance(var, LocalIRVariable) and var.is_storage:
                    for refer_to in var.refers_to:
                        # the following should always be true
                        if isinstance(refer_to, (StateIRVariable, LocalIRVariable)):
                            self._ssa_vars_read.append(refer_to)

        elif isinstance(ir, Index):
            variable_right: RVALUE = ir.variable_right
            if isinstance(variable_right, (StateIRVariable, LocalIRVariable)):
                self._ssa_vars_read.append(variable_right)

            if isinstance(variable_right, ReferenceVariable):
                origin = variable_right.points_to_origin
                if isinstance(origin, (StateIRVariable, LocalIRVariable)):
                    self._ssa_vars_read.append(origin)

    def _update_write_using_ssa(self, ir: Operation) -> None:
        """
        Update self._ssa_vars_written
        This look for all operations that write a IRvariable
        It uses the result of the storage pointer

        Index/member/Length are not considering writing to anything
        For index/member it is implictely handled when their associated RefernceVarible are written

        """

        if isinstance(ir, OperationWithLValue) and not isinstance(ir, Phi):
            if isinstance(ir, (Index, Member, Length)):
                return  # Don't consider Member and Index operations -> ReferenceVariable

            var = ir.lvalue

            if isinstance(var, ReferenceVariable):
                var = var.points_to_origin

            candidates = [var]

            # If we write to a storage pointer, add everything it points to as target
            # if it's a variable declaration we do not want to consider the right variable written in that case
            # string storage ss = s; // s is a storage variable but should not be considered written at that point
            if (
                isinstance(var, LocalIRVariable)
                and var.is_storage
                and ir.node.type is not NodeType.VARIABLE
            ):
                candidates += var.refers_to

            for var in candidates:
                # Only store non-slithIR variables
                if var and isinstance(var, (StateIRVariable, LocalIRVariable)):
                    if isinstance(ir, PhiCallback):
                        continue
                    self._ssa_vars_written.append(var)

    def update_read_write_using_ssa(self) -> None:
        for ir in self.irs_ssa:
            if isinstance(ir, PhiCallback):
                continue

            self._update_read_using_ssa(ir)
            self._update_write_using_ssa(ir)

        self._ssa_vars_read = list(set(self._ssa_vars_read))
        self._ssa_state_vars_read = [v for v in self._ssa_vars_read if isinstance(v, StateVariable)]
        self._ssa_local_vars_read = [v for v in self._ssa_vars_read if isinstance(v, LocalVariable)]
        self._ssa_vars_written = list(set(self._ssa_vars_written))
        self._ssa_state_vars_written = [
            v for v in self._ssa_vars_written if v and isinstance(v, StateIRVariable)
        ]
        self._ssa_local_vars_written = [
            v for v in self._ssa_vars_written if v and isinstance(v, LocalIRVariable)
        ]

        vars_read = [self._convert_ssa(x) for x in self._ssa_vars_read]
        vars_written = [self._convert_ssa(x) for x in self._ssa_vars_written]

        self._vars_read += [v_ for v_ in vars_read if v_ and v_ not in self._vars_read]
        self._state_vars_read = [v for v in self._vars_read if isinstance(v, StateVariable)]
        self._local_vars_read = [v for v in self._vars_read if isinstance(v, LocalVariable)]

        self._vars_written += [v_ for v_ in vars_written if v_ and v_ not in self._vars_written]
        self._state_vars_written = [v for v in self._vars_written if isinstance(v, StateVariable)]
        self._local_vars_written = [v for v in self._vars_written if isinstance(v, LocalVariable)]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Built in definitions
    ###################################################################################
    ###################################################################################

    def __str__(self) -> str:
        additional_info = ""
        if self.expression:
            additional_info += " " + str(self.expression)
        elif self.variable_declaration:
            additional_info += " " + str(self.variable_declaration)
        txt = str(self._node_type.value) + additional_info
        return txt


# endregion
###################################################################################
###################################################################################
# region Utils
###################################################################################
###################################################################################


def link_nodes(node1: Node, node2: Node) -> None:
    node1.add_son(node2)
    node2.add_father(node1)


def insert_node(origin: Node, node_inserted: Node) -> None:
    sons = origin.sons
    link_nodes(origin, node_inserted)
    for son in sons:
        son.remove_father(origin)
        origin.remove_son(son)

        link_nodes(node_inserted, son)


def recheable(node: Node) -> set[Node]:
    """
    Return the set of nodes reacheable from the node
    :param node:
    :return: set(Node)
    """
    nodes = node.sons
    visited = set()
    while nodes:
        next_node = nodes[0]
        nodes = nodes[1:]
        if next_node not in visited:
            visited.add(next_node)
            for son in next_node.sons:
                if son not in visited:
                    nodes.append(son)
    return visited


# endregion



import logging
import re
from typing import TYPE_CHECKING, Union
from collections.abc import ValuesView

from slither.core.declarations.custom_error_contract import CustomErrorContract
from slither.core.declarations.custom_error_top_level import CustomErrorTopLevel
from slither.core.declarations.function_contract import FunctionContract
from slither.core.expressions.literal import Literal
from slither.core.solidity_types import TypeAlias, TypeAliasTopLevel, TypeAliasContract
from slither.core.solidity_types.array_type import ArrayType
from slither.core.solidity_types.elementary_type import (
    ElementaryType,
    ElementaryTypeName,
)
from slither.core.solidity_types.function_type import FunctionType
from slither.core.solidity_types.mapping_type import MappingType
from slither.core.solidity_types.type import Type
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.core.variables.function_type_variable import FunctionTypeVariable
from slither.exceptions import SlitherError
from slither.solc_parsing.exceptions import ParsingError
from slither.solc_parsing.expressions.expression_parsing import CallerContextExpression

if TYPE_CHECKING:
    from slither.core.declarations import Structure, Enum, Function
    from slither.core.declarations.contract import Contract
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc

logger = logging.getLogger("TypeParsing")


class UnknownType:
    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name


def _find_from_type_name(
    name: str,
    functions_direct_access: list["Function"],
    contracts_direct_access: list["Contract"],
    structures_direct_access: list["Structure"],
    all_structures: ValuesView["Structure"],
    enums_direct_access: list["Enum"],
    all_enums: ValuesView["Enum"],
) -> Type:
    name_elementary = name.split(" ")[0]
    if "[" in name_elementary:
        name_elementary = name_elementary[0 : name_elementary.find("[")]
    if name_elementary in ElementaryTypeName:
        depth = name.count("[")
        if depth:
            return ArrayType(ElementaryType(name_elementary), Literal(depth, "uint256"))
        return ElementaryType(name_elementary)
    # We first look for contract
    # To avoid collision
    # Ex: a structure with the name of a contract
    name_contract = name
    if name_contract.startswith("contract "):
        name_contract = name_contract[len("contract ") :]
    if name_contract.startswith("library "):
        name_contract = name_contract[len("library ") :]
    var_type = next((c for c in contracts_direct_access if c.name == name_contract), None)

    if not var_type:
        var_type = next((st for st in structures_direct_access if st.name == name), None)
    if not var_type:
        var_type = next((e for e in enums_direct_access if e.name == name), None)
    if not var_type:
        # any contract can refer to another contract's enum
        enum_name = name
        if enum_name.startswith("enum "):
            enum_name = enum_name[len("enum ") :]
        elif enum_name.startswith("type(enum"):
            enum_name = enum_name[len("type(enum ") : -1]
        # all_enums = [c.enums for c in contracts]
        # all_enums = [item for sublist in all_enums for item in sublist]
        # all_enums += contract.slither.enums_top_level
        var_type = next((e for e in all_enums if e.canonical_name == enum_name), None)
        if not var_type:
            var_type = next((e for e in all_enums if e.name == enum_name), None)
    if not var_type:
        # any contract can refer to another contract's structure
        name_struct = name
        if name_struct.startswith("struct "):
            name_struct = name_struct[len("struct ") :]
            name_struct = name_struct.split(" ")[0]  # remove stuff like storage pointer at the end
        # all_structures = [c.structures for c in contracts]
        # all_structures = [item for sublist in all_structures for item in sublist]
        # all_structures += contract.slither.structures_top_level
        var_type = next((st for st in all_structures if st.canonical_name == name_struct), None)
        if not var_type:
            var_type = next((st for st in all_structures if st.name == name_struct), None)
        # case where struct xxx.xx[] where not well formed in the AST
        if not var_type:
            depth = 0
            while name_struct.endswith("[]"):
                name_struct = name_struct[0:-2]
                depth += 1
            var_type = next((st for st in all_structures if st.canonical_name == name_struct), None)
            if var_type:
                return ArrayType(UserDefinedType(var_type), Literal(depth, "uint256"))

    if not var_type:
        var_type = next((f for f in functions_direct_access if f.name == name), None)
    if not var_type:
        if name.startswith("function "):
            found = re.findall(
                r"function \(([ ()\[\]a-zA-Z0-9\.,]*?)\)(?: payable)?(?: (?:external|internal|pure|view))?(?: returns \(([a-zA-Z0-9() \.,]*)\))?",
                name,
            )
            assert len(found) == 1
            params = [v for v in found[0][0].split(",") if v != ""]
            return_values = (
                [v for v in found[0][1].split(",") if v != ""] if len(found[0]) > 1 else []
            )
            params = [
                _find_from_type_name(
                    p,
                    functions_direct_access,
                    contracts_direct_access,
                    structures_direct_access,
                    all_structures,
                    enums_direct_access,
                    all_enums,
                )
                for p in params
            ]
            return_values = [
                _find_from_type_name(
                    r,
                    functions_direct_access,
                    contracts_direct_access,
                    structures_direct_access,
                    all_structures,
                    enums_direct_access,
                    all_enums,
                )
                for r in return_values
            ]
            params_vars = []
            return_vars = []
            for p in params:
                var = FunctionTypeVariable()
                var.set_type(p)
                params_vars.append(var)
            for r in return_values:
                var = FunctionTypeVariable()
                var.set_type(r)
                return_vars.append(var)
            return FunctionType(params_vars, return_vars)
    if not var_type:
        if name.startswith("mapping("):
            # nested mapping declared with var
            if name.count("mapping(") == 1:
                found = re.findall(r"mapping\(([a-zA-Z0-9\.]*) => ([ a-zA-Z0-9\.\[\]]*)\)", name)
            else:
                found = re.findall(
                    r"mapping\(([a-zA-Z0-9\.]*) => (mapping\([=> a-zA-Z0-9\.\[\]]*\))\)",
                    name,
                )
            assert len(found) == 1
            from_ = found[0][0]
            to_ = found[0][1]

            from_type = _find_from_type_name(
                from_,
                functions_direct_access,
                contracts_direct_access,
                structures_direct_access,
                all_structures,
                enums_direct_access,
                all_enums,
            )
            to_type = _find_from_type_name(
                to_,
                functions_direct_access,
                contracts_direct_access,
                structures_direct_access,
                all_structures,
                enums_direct_access,
                all_enums,
            )

            return MappingType(from_type, to_type)

    if not var_type:
        raise ParsingError("Type not found " + str(name))
    return UserDefinedType(var_type)


def _add_type_references(type_found: Type, src: str, sl: "SlitherCompilationUnit") -> None:
    if isinstance(type_found, UserDefinedType):
        type_found.type.add_reference_from_raw_source(src, sl)
    elif isinstance(type_found, (TypeAliasTopLevel, TypeAliasContract)):
        type_found.type.add_reference_from_raw_source(src, sl)
        type_found.add_reference_from_raw_source(src, sl)


# TODO: since the add of FileScope, we can probably refactor this function and makes it a lot simpler
def parse_type(
    t: dict | UnknownType,
    caller_context: Union[CallerContextExpression, "SlitherCompilationUnitSolc"],
) -> Type:
    """
    caller_context can be a SlitherCompilationUnitSolc because we recursively call the function
    and go up in the context's scope. If we are really lost we just go over the SlitherCompilationUnitSolc

    :param t:
    :type t:
    :param caller_context:
    :type caller_context:
    :return:
    :rtype:
    """
    # local import to avoid circular dependency

    from slither.solc_parsing.expressions.expression_parsing import parse_expression
    from slither.solc_parsing.variables.function_type_variable import FunctionTypeVariableSolc
    from slither.solc_parsing.declarations.contract import ContractSolc
    from slither.solc_parsing.declarations.function import FunctionSolc
    from slither.solc_parsing.declarations.using_for_top_level import UsingForTopLevelSolc
    from slither.solc_parsing.declarations.custom_error import CustomErrorSolc
    from slither.solc_parsing.declarations.structure_top_level import StructureTopLevelSolc
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
    from slither.solc_parsing.variables.top_level_variable import TopLevelVariableSolc
    from slither.solc_parsing.declarations.event_top_level import EventTopLevelSolc

    sl: SlitherCompilationUnit
    renaming: dict[str, str]
    type_aliases: dict[str, TypeAlias]
    enums_direct_access: list[Enum] = []
    # Note: for convenience top level functions use the same parser as function in contract
    # but contract_parser is set to None
    if isinstance(caller_context, SlitherCompilationUnitSolc) or (
        isinstance(caller_context, FunctionSolc) and caller_context.contract_parser is None
    ):
        structures_direct_access: list[Structure]
        if isinstance(caller_context, SlitherCompilationUnitSolc):
            sl = caller_context.compilation_unit
            next_context = caller_context
            renaming = {}
            type_aliases = sl.type_aliases
        else:
            assert isinstance(caller_context, FunctionSolc)
            sl = caller_context.underlying_function.compilation_unit
            next_context = caller_context.slither_parser
            renaming = caller_context.underlying_function.file_scope.renaming
            type_aliases = caller_context.underlying_function.file_scope.type_aliases
        structures_direct_access = sl.structures_top_level
        all_structuress = [c.structures for c in sl.contracts]
        all_structures = [item for sublist in all_structuress for item in sublist]
        all_structures += structures_direct_access
        enums_direct_access += sl.enums_top_level
        all_enumss = [c.enums for c in sl.contracts]
        all_enums = [item for sublist in all_enumss for item in sublist]
        all_enums += enums_direct_access
        contracts = sl.contracts
        functions = []
    elif isinstance(
        caller_context,
        (
            StructureTopLevelSolc,
            CustomErrorSolc,
            TopLevelVariableSolc,
            UsingForTopLevelSolc,
            EventTopLevelSolc,
        ),
    ):
        if isinstance(caller_context, StructureTopLevelSolc):
            scope = caller_context.underlying_structure.file_scope
        elif isinstance(caller_context, TopLevelVariableSolc):
            scope = caller_context.underlying_variable.file_scope
        elif isinstance(caller_context, UsingForTopLevelSolc):
            scope = caller_context.underlying_using_for.file_scope
        elif isinstance(caller_context, EventTopLevelSolc):
            scope = caller_context.underlying_event.file_scope
        else:
            assert isinstance(caller_context, CustomErrorSolc)
            custom_error = caller_context.underlying_custom_error
            if isinstance(custom_error, CustomErrorTopLevel):
                scope = custom_error.file_scope
            else:
                assert isinstance(custom_error, CustomErrorContract)
                scope = custom_error.contract.file_scope

        sl = caller_context.compilation_unit
        next_context = caller_context
        structures_direct_access = list(scope.structures.values())
        all_structuress = [c.structures for c in scope.contracts.values()]
        all_structures = [item for sublist in all_structuress for item in sublist]
        all_structures += structures_direct_access

        enums_direct_access = []
        all_enumss = [c.enums for c in scope.contracts.values()]
        all_enums = [item for sublist in all_enumss for item in sublist]
        all_enums += scope.enums.values()

        contracts = scope.contracts.values()
        functions = list(scope.functions)

        renaming = scope.renaming
        type_aliases = scope.type_aliases
    elif isinstance(caller_context, (ContractSolc, FunctionSolc)):
        sl = caller_context.compilation_unit
        if isinstance(caller_context, FunctionSolc):
            underlying_func = caller_context.underlying_function
            # If contract_parser is set to None, then underlying_function is a FunctionContract
            # See note above
            assert isinstance(underlying_func, FunctionContract)
            contract = underlying_func.contract
            next_context = caller_context.contract_parser
            scope = underlying_func.file_scope
        else:
            contract = caller_context.underlying_contract
            next_context = caller_context
            scope = contract.file_scope

        structures_direct_access = contract.structures
        structures_direct_access += scope.structures.values()
        all_structuress = [c.structures for c in scope.contracts.values()]
        all_structures = [item for sublist in all_structuress for item in sublist]
        all_structures += scope.structures.values()
        enums_direct_access += contract.enums
        enums_direct_access += scope.enums.values()
        all_enumss = [c.enums for c in scope.contracts.values()]
        all_enums = [item for sublist in all_enumss for item in sublist]
        all_enums += scope.enums.values()
        contracts = scope.contracts.values()
        functions = contract.functions + contract.modifiers

        renaming = scope.renaming
        type_aliases = scope.type_aliases
    else:
        raise ParsingError(f"Incorrect caller context: {type(caller_context)}")

    is_compact_ast = caller_context.is_compact_ast
    if is_compact_ast:
        key = "nodeType"
    else:
        key = "name"

    if isinstance(t, UnknownType):
        name = t.name
        if name in renaming:
            name = renaming[name]
        if name in type_aliases:
            return type_aliases[name]
        return _find_from_type_name(
            name,
            functions,
            contracts,
            structures_direct_access,
            all_structures,
            enums_direct_access,
            all_enums,
        )

    if t[key] == "ElementaryTypeName":
        if is_compact_ast:
            return ElementaryType(t["name"])
        return ElementaryType(t["attributes"][key])

    if t[key] == "UserDefinedTypeName":
        if is_compact_ast:
            name = t["typeDescriptions"]["typeString"]
            if name in renaming:
                name = renaming[name]
            if name in type_aliases:
                _add_type_references(type_aliases[name], t["src"], sl)
                return type_aliases[name]
            type_found = _find_from_type_name(
                name,
                functions,
                contracts,
                structures_direct_access,
                all_structures,
                enums_direct_access,
                all_enums,
            )
            _add_type_references(type_found, t["src"], sl)
            return type_found

        # Determine if we have a type node (otherwise we use the name node, as some older solc did not have 'type').
        type_name_key = "type" if "type" in t["attributes"] else key

        name = t["attributes"][type_name_key]
        if name in renaming:
            name = renaming[name]
        if name in type_aliases:
            _add_type_references(type_aliases[name], t["src"], sl)
            return type_aliases[name]
        type_found = _find_from_type_name(
            name,
            functions,
            contracts,
            structures_direct_access,
            all_structures,
            enums_direct_access,
            all_enums,
        )
        _add_type_references(type_found, t["src"], sl)
        return type_found

    # Introduced with Solidity 0.8
    if t[key] == "IdentifierPath":
        if is_compact_ast:
            name = t["name"]
            if name in renaming:
                name = renaming[name]
            if name in type_aliases:
                return type_aliases[name]
            type_found = _find_from_type_name(
                name,
                functions,
                contracts,
                structures_direct_access,
                all_structures,
                enums_direct_access,
                all_enums,
            )
            _add_type_references(type_found, t["src"], sl)
            return type_found

        raise SlitherError("Solidity 0.8 not supported with the legacy AST")

    if t[key] == "ArrayTypeName":
        length = None
        if is_compact_ast:
            if t.get("length", None):
                length = parse_expression(t["length"], caller_context)
            array_type = parse_type(t["baseType"], next_context)
        else:
            if len(t["children"]) == 2:
                length = parse_expression(t["children"][1], caller_context)
            else:
                assert len(t["children"]) == 1
            array_type = parse_type(t["children"][0], next_context)
        return ArrayType(array_type, length)

    if t[key] == "Mapping":
        if is_compact_ast:
            mappingFrom = parse_type(t["keyType"], next_context)
            mappingTo = parse_type(t["valueType"], next_context)
        else:
            assert len(t["children"]) == 2

            mappingFrom = parse_type(t["children"][0], next_context)
            mappingTo = parse_type(t["children"][1], next_context)

        return MappingType(mappingFrom, mappingTo)

    if t[key] == "FunctionTypeName":
        if is_compact_ast:
            params = t["parameterTypes"]
            return_values = t["returnParameterTypes"]
            index = "parameters"
        else:
            assert len(t["children"]) == 2
            params = t["children"][0]
            return_values = t["children"][1]
            index = "children"

        assert params[key] == "ParameterList"
        assert return_values[key] == "ParameterList"

        params_vars: list[FunctionTypeVariable] = []
        return_values_vars: list[FunctionTypeVariable] = []
        for p in params[index]:
            var = FunctionTypeVariable()
            var.set_offset(p["src"], caller_context.compilation_unit)

            var_parser = FunctionTypeVariableSolc(var, p)
            var_parser.analyze(caller_context)

            params_vars.append(var)
        for p in return_values[index]:
            var = FunctionTypeVariable()
            var.set_offset(p["src"], caller_context.compilation_unit)

            var_parser = FunctionTypeVariableSolc(var, p)
            var_parser.analyze(caller_context)

            return_values_vars.append(var)

        return FunctionType(params_vars, return_values_vars)

    raise ParsingError(f"Type name not found {(t)} in {scope.filename}")



import argparse
import logging
import os
import sys
import traceback

from slither.tools.similarity.cache import save_cache
from slither.tools.similarity.encode import encode_contract, load_contracts
from slither.tools.similarity.model import train_unsupervised

logger = logging.getLogger("Slither-simil")


def train(args: argparse.Namespace) -> None:
    try:
        last_data_train_filename = "last_data_train.txt"
        model_filename = args.model
        dirname = args.input

        if dirname is None:
            logger.error("The train mode requires the input parameter.")
            sys.exit(-1)

        contracts = load_contracts(dirname, **vars(args))
        logger.info("Saving extracted data into %s", last_data_train_filename)
        cache = []
        with open(last_data_train_filename, "w", encoding="utf8") as f:
            for filename in contracts:
                # cache[filename] = dict()
                for (filename_inner, contract, function), ir in encode_contract(
                    filename, **vars(args)
                ).items():
                    if ir != []:
                        x = " ".join(ir)
                        f.write(x + "\n")
                        cache.append((os.path.split(filename_inner)[-1], contract, function, x))

        logger.info("Starting training")
        model = train_unsupervised(input=last_data_train_filename, model="skipgram")
        logger.info("Training complete")
        logger.info("Saving model")
        model.save_model(model_filename)

        for i, (filename, contract, function, irs) in enumerate(cache):
            cache[i] = ((filename, contract, function), model.get_sentence_vector(irs))

        logger.info("Saving cache in cache.npz")
        save_cache(cache, "cache.npz")
        logger.info("Done!")

    except Exception:
        logger.error(f"Error in {args.filename}")
        logger.error(traceback.format_exc())
        sys.exit(-1)



from slither.tools.properties.properties.properties import (
    Property,
    PropertyType,
    PropertyReturn,
    PropertyCaller,
)

ERC20_NotBurnable = [
    Property(
        name="crytic_supply_constant_ERC20PropertiesNotBurnable()",
        description="The total supply does not decrease.",
        content="""
\t\treturn initialTotalSupply == this.totalSupply();""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ANY,
    ),
]


# Require burn(address) returns()
ERC20_Burnable = [
    Property(
        name="crytic_supply_constant_ERC20PropertiesNotBurnable()",
        description="Cannot burn more than available balance",
        content="""
\t\tuint balance = balanceOf(msg.sender);
\t\tburn(balance + 1);
\t\treturn false;""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.THROW,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    )
]



import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc


class CallerContextExpression(metaclass=abc.ABCMeta):
    """
    This class is inherited by all the declarations class that can be used in the expression/type parsing
    As a source of context/scope

    It is used by any declaration class that can be top-level and require complex parsing

    """

    @property
    @abc.abstractmethod
    def is_compact_ast(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def compilation_unit(self) -> "SlitherCompilationUnit":
        pass

    @abc.abstractmethod
    def get_key(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        pass



import re
import json
from pathlib import Path

import pytest
from deepdiff import DeepDiff
from web3.contract import Contract

from slither import Slither
from slither.tools.read_storage import SlitherReadStorage, RpcInfo

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def get_source_file(file_path) -> str:
    with open(file_path, encoding="utf8") as f:
        source = f.read()

    return source


def deploy_contract(w3, ganache, contract_bin, contract_abi) -> Contract:
    """Deploy contract to the local ganache network"""
    signed_txn = w3.eth.account.sign_transaction(
        {
            "nonce": w3.eth.get_transaction_count(ganache.eth_address),
            "maxFeePerGas": 20000000000,
            "maxPriorityFeePerGas": 1,
            "gas": 15000000,
            "to": b"",
            "data": "0x" + contract_bin,
            "chainId": 1,
        },
        ganache.eth_privkey,
    )
    tx_hash = w3.eth.send_raw_transaction(signed_txn.raw_transaction)
    address = w3.eth.get_transaction_receipt(tx_hash)["contractAddress"]
    contract = w3.eth.contract(address, abi=contract_abi)
    return contract


@pytest.mark.parametrize(
    "test_contract, storage_file",
    [("StorageLayout", "storage_layout"), ("UnstructuredStorageLayout", "unstructured_storage")],
)
@pytest.mark.usefixtures("web3", "ganache")
def test_read_storage(test_contract, storage_file, web3, ganache, solc_binary_path) -> None:
    solc_path = solc_binary_path(version="0.8.10")

    assert web3.is_connected()
    bin_path = Path(TEST_DATA_DIR, f"{test_contract}.bin").as_posix()
    abi_path = Path(TEST_DATA_DIR, f"{test_contract}.abi").as_posix()
    bytecode = get_source_file(bin_path)
    abi = get_source_file(abi_path)
    contract = deploy_contract(web3, ganache, bytecode, abi)
    contract.functions.store().transact({"from": ganache.eth_address})
    address = contract.address

    sl = Slither(Path(TEST_DATA_DIR, f"{test_contract}.sol").as_posix(), solc=solc_path)
    contracts = sl.contracts

    rpc_info: RpcInfo = RpcInfo(ganache.provider)
    srs = SlitherReadStorage(contracts, 100, rpc_info)
    srs.unstructured = True
    srs.storage_address = address
    srs.get_all_storage_variables()
    srs.get_storage_layout()
    srs.walk_slot_info(srs.get_slot_values)
    actual_file = Path(TEST_DATA_DIR, "storage_layout.json").as_posix()
    with open(actual_file, "w", encoding="utf-8") as file:
        slot_infos_json = srs.to_json()
        json.dump(slot_infos_json, file, indent=4)

    expected_file = Path(TEST_DATA_DIR, f"TEST_{storage_file}.json").as_posix()

    with open(expected_file, encoding="utf8") as f:
        expected = json.load(f)
    with open(actual_file, encoding="utf8") as f:
        actual = json.load(f)

    diff = DeepDiff(expected, actual, ignore_order=True, verbose_level=2, view="tree")
    if diff:
        for change in diff.get("values_changed", []):
            path_list = re.findall(r"\['(.*?)'\]", change.path())
            path = "_".join(path_list)
            with open(f"{path}_expected.txt", "w", encoding="utf8") as f:
                f.write(str(change.t1))
            with open(f"{path}_actual.txt", "w", encoding="utf8") as f:
                f.write(str(change.t2))

    assert not diff



from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.statements.pyth_unchecked import PythUnchecked


class PythUncheckedPublishTime(PythUnchecked):
    """
    Documentation: This detector finds when the publishTime of a Pyth price is not checked
    """

    ARGUMENT = "pyth-unchecked-publishtime"
    HELP = "Detect when the publishTime of a Pyth price is not checked"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#pyth-unchecked-publishtime"
    )
    WIKI_TITLE = "Pyth unchecked publishTime"
    WIKI_DESCRIPTION = "Detect when the publishTime of a Pyth price is not checked"
    WIKI_RECOMMENDATION = "Check the publishTime of a Pyth price."

    WIKI_EXPLOIT_SCENARIO = """
```solidity
import "@pythnetwork/pyth-sdk-solidity/IPyth.sol";
import "@pythnetwork/pyth-sdk-solidity/PythStructs.sol";

contract C {
    IPyth pyth;

    constructor(IPyth _pyth) {
        pyth = _pyth;
    }

    function bad(bytes32 id) public {
        PythStructs.Price memory price = pyth.getEmaPriceUnsafe(id);
        // Use price
    }
}
```
The function `A` uses the price without checking its `publishTime` coming from the `getEmaPriceUnsafe` function.
"""

    PYTH_FUNCTIONS = [
        "getEmaPrice",
        # "getEmaPriceNoOlderThan",
        "getEmaPriceUnsafe",
        "getPrice",
        # "getPriceNoOlderThan",
        "getPriceUnsafe",
    ]

    PYTH_FIELD = "publishTime"



import logging
import re
from typing import Union, TYPE_CHECKING, Any

import slither.core.expressions.type_conversion
from slither.core.declarations.solidity_variables import (
    SOLIDITY_VARIABLES_COMPOSED,
    SolidityVariableComposed,
)
from slither.core.expressions import (
    CallExpression,
    ConditionalExpression,
    ElementaryTypeNameExpression,
    Identifier,
    IndexAccess,
    Literal,
    MemberAccess,
    NewArray,
    NewContract,
    NewElementaryType,
    SuperCallExpression,
    SuperIdentifier,
    TupleExpression,
    TypeConversion,
    UnaryOperation,
    UnaryOperationType,
)
from slither.core.expressions.assignment_operation import (
    AssignmentOperation,
    AssignmentOperationType,
)
from slither.core.expressions.binary_operation import (
    BinaryOperation,
    BinaryOperationType,
)
from slither.core.solidity_types import (
    ArrayType,
    ElementaryType,
    UserDefinedType,
)
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.exceptions import ParsingError, VariableNotFound
from slither.core.variables.variable import Variable
from slither.solc_parsing.expressions.find_variable import find_variable
from slither.solc_parsing.solidity_types.type_parsing import UnknownType, parse_type


if TYPE_CHECKING:
    from slither.core.expressions.expression import Expression
    from slither.solc_parsing.declarations.contract import ContractSolc
    from slither.solc_parsing.declarations.function import FunctionSolc
    from slither.solc_parsing.variables.top_level_variable import TopLevelVariableSolc

logger = logging.getLogger("ExpressionParsing")


# region Filtering
###################################################################################
###################################################################################


def filter_name(value: str) -> str:
    value = value.replace(" memory", "")
    value = value.replace(" storage", "")
    value = value.replace(" external", "")
    value = value.replace(" internal", "")
    value = value.replace("struct ", "")
    value = value.replace("contract ", "")
    value = value.replace("enum ", "")
    value = value.replace(" ref", "")
    value = value.replace(" pointer", "")
    value = value.replace(" pure", "")
    value = value.replace(" view", "")
    value = value.replace(" constant", "")
    value = value.replace(" payable", "")
    value = value.replace("function (", "function(")
    value = value.replace("returns (", "returns(")
    value = value.replace(" calldata", "")

    # remove the text remaining after functio(...)
    # which should only be ..returns(...)
    # nested parenthesis so we use a system of counter on parenthesis
    idx = value.find("(")
    if idx:
        counter = 1
        max_idx = len(value)
        while counter:
            assert idx < max_idx
            idx = idx + 1
            if value[idx] == "(":
                counter += 1
            elif value[idx] == ")":
                counter -= 1
        value = value[: idx + 1]
    return value


# endregion

###################################################################################
###################################################################################
# region Parsing
###################################################################################
###################################################################################


def parse_call(
    expression: dict, caller_context: Union["FunctionSolc", "ContractSolc", "TopLevelVariableSolc"]
) -> (
    slither.core.expressions.call_expression.CallExpression
    | slither.core.expressions.type_conversion.TypeConversion
):
    src = expression["src"]
    if caller_context.is_compact_ast:
        attributes = expression
        type_conversion = expression["kind"] == "typeConversion"
        type_return = attributes["typeDescriptions"]["typeString"]

    else:
        attributes = expression["attributes"]
        type_conversion = attributes["type_conversion"]
        type_return = attributes["type"]

    if type_conversion:
        type_call = parse_type(UnknownType(type_return), caller_context)
        if caller_context.is_compact_ast:
            assert len(expression["arguments"]) == 1
            expression_to_parse = expression["arguments"][0]
        else:
            children = expression["children"]
            assert len(children) == 2
            type_info = children[0]
            expression_to_parse = children[1]
            assert type_info["name"] in [
                "ElementaryTypenameExpression",
                "ElementaryTypeNameExpression",
                "Identifier",
                "TupleExpression",
                "IndexAccess",
                "MemberAccess",
            ]

        expression = parse_expression(expression_to_parse, caller_context)
        t = TypeConversion(expression, type_call)
        t.set_offset(src, caller_context.compilation_unit)
        if isinstance(type_call, UserDefinedType):
            type_call.type.references.append(t.source_mapping)
        return t

    call_gas = None
    call_value = None
    call_salt = None
    if caller_context.is_compact_ast:
        called = parse_expression(expression["expression"], caller_context)
        # If the next expression is a FunctionCallOptions
        # We can here the gas/value information
        # This is only available if the syntax is {gas: , value: }
        # For the .gas().value(), the member are considered as function call
        # And converted later to the correct info (convert.py)
        if expression["expression"][caller_context.get_key()] == "FunctionCallOptions":
            call_with_options = expression["expression"]
            for idx, name in enumerate(call_with_options.get("names", [])):
                option = parse_expression(call_with_options["options"][idx], caller_context)
                if name == "value":
                    call_value = option
                if name == "gas":
                    call_gas = option
                if name == "salt":
                    call_salt = option
        arguments = []
        if expression["arguments"]:
            arguments = [parse_expression(a, caller_context) for a in expression["arguments"]]
    else:
        children = expression["children"]
        called = parse_expression(children[0], caller_context)
        arguments = [parse_expression(a, caller_context) for a in children[1::]]

    if isinstance(called, SuperIdentifier):
        sp = SuperCallExpression(called, arguments, type_return)
        sp.set_offset(expression["src"], caller_context.compilation_unit)
        return sp
    names = expression["names"] if "names" in expression and len(expression["names"]) > 0 else None
    call_expression = CallExpression(called, arguments, type_return, names=names)
    call_expression.set_offset(src, caller_context.compilation_unit)

    # Only available if the syntax {gas:, value:} was used
    call_expression.call_gas = call_gas
    call_expression.call_value = call_value
    call_expression.call_salt = call_salt
    return call_expression


def parse_super_name(expression: dict, is_compact_ast: bool) -> str:
    if is_compact_ast:
        assert expression["nodeType"] == "MemberAccess"
        base_name = expression["memberName"]
        arguments = expression["typeDescriptions"]["typeString"]
    else:
        assert expression["name"] == "MemberAccess"
        attributes = expression["attributes"]
        base_name = attributes["member_name"]
        arguments = attributes["type"]

    assert arguments.startswith("function ")
    # remove function (...()
    arguments = arguments[len("function ") :]

    arguments = filter_name(arguments)
    if " " in arguments:
        arguments = arguments[: arguments.find(" ")]

    return base_name + arguments


def _parse_elementary_type_name_expression(
    expression: dict, is_compact_ast: bool, caller_context: CallerContextExpression
) -> ElementaryTypeNameExpression:
    # nop exression
    # uint;
    if is_compact_ast:
        value = expression["typeName"]
    else:
        if "children" in expression:
            value = expression["children"][0]["attributes"]["name"]
        else:
            value = expression["attributes"]["value"]
    if isinstance(value, dict):
        t = parse_type(value, caller_context)
    else:
        t = parse_type(UnknownType(value), caller_context)
    e = ElementaryTypeNameExpression(t)
    e.set_offset(expression["src"], caller_context.compilation_unit)
    return e


def _user_defined_op_call(
    caller_context: CallerContextExpression, src, function_id: int, args: list[Any], type_call: str
) -> CallExpression:
    var, was_created = find_variable(None, caller_context, function_id)

    if was_created:
        var.set_offset(src, caller_context.compilation_unit)

    identifier = Identifier(var)
    identifier.set_offset(src, caller_context.compilation_unit)

    var.references.append(identifier.source_mapping)

    call = CallExpression(identifier, args, type_call)
    call.set_offset(src, caller_context.compilation_unit)
    return call


def parse_expression(expression: dict, caller_context: CallerContextExpression) -> "Expression":
    """

    Returns:
        str: expression
    """
    #  Expression
    #    = Expression ('++' | '--')
    #    | NewExpression
    #    | IndexAccess
    #    | MemberAccess
    #    | FunctionCall
    #    | '(' Expression ')'
    #    | ('!' | '~' | 'delete' | '++' | '--' | '+' | '-') Expression
    #    | Expression '**' Expression
    #    | Expression ('*' | '/' | '%') Expression
    #    | Expression ('+' | '-') Expression
    #    | Expression ('<<' | '>>') Expression
    #    | Expression '&' Expression
    #    | Expression '^' Expression
    #    | Expression '|' Expression
    #    | Expression ('<' | '>' | '<=' | '>=') Expression
    #    | Expression ('==' | '!=') Expression
    #    | Expression '&&' Expression
    #    | Expression '||' Expression
    #    | Expression '?' Expression ':' Expression
    #    | Expression ('=' | '|=' | '^=' | '&=' | '<<=' | '>>=' | '+=' | '-=' | '*=' | '/=' | '%=') Expression
    #    | PrimaryExpression
    # The AST naming does not follow the spec
    assert isinstance(caller_context, CallerContextExpression)
    name = expression[caller_context.get_key()]
    is_compact_ast = caller_context.is_compact_ast
    src = expression["src"]

    if name == "UnaryOperation":
        if is_compact_ast:
            attributes = expression
            expression = parse_expression(expression["subExpression"], caller_context)
        else:
            attributes = expression["attributes"]
            assert len(expression["children"]) == 1
            expression = parse_expression(expression["children"][0], caller_context)
        assert "prefix" in attributes

        # Use of user defined operation
        if "function" in attributes:
            return _user_defined_op_call(
                caller_context,
                src,
                attributes["function"],
                [expression],
                attributes["typeDescriptions"]["typeString"],
            )

        operation_type = UnaryOperationType.get_type(attributes["operator"], attributes["prefix"])
        unary_op = UnaryOperation(expression, operation_type)
        unary_op.set_offset(src, caller_context.compilation_unit)
        return unary_op

    if name == "BinaryOperation":
        if is_compact_ast:
            attributes = expression
            left_expression = parse_expression(expression["leftExpression"], caller_context)
            right_expression = parse_expression(expression["rightExpression"], caller_context)
        else:
            assert len(expression["children"]) == 2
            attributes = expression["attributes"]
            left_expression = parse_expression(expression["children"][0], caller_context)
            right_expression = parse_expression(expression["children"][1], caller_context)

        # Use of user defined operation
        if "function" in attributes:
            return _user_defined_op_call(
                caller_context,
                src,
                attributes["function"],
                [left_expression, right_expression],
                attributes["typeDescriptions"]["typeString"],
            )

        operation_type = BinaryOperationType.get_type(attributes["operator"])
        binary_op = BinaryOperation(left_expression, right_expression, operation_type)
        binary_op.set_offset(src, caller_context.compilation_unit)
        return binary_op

    if name in "FunctionCall":
        return parse_call(expression, caller_context)

    if name == "FunctionCallOptions":
        # call/gas info are handled in parse_call
        if is_compact_ast:
            called = parse_expression(expression["expression"], caller_context)
        else:
            called = parse_expression(expression["children"][0], caller_context)
        assert isinstance(called, (MemberAccess, NewContract, Identifier, TupleExpression))
        return called

    if name == "TupleExpression":
        #     For expression like
        #     (a,,c) = (1,2,3)
        #     the AST provides only two children in the left side
        #     We check the type provided (tuple(uint256,,uint256))
        #     To determine that there is an empty variable
        #     Otherwhise we would not be able to determine that
        #     a = 1, c = 3, and 2 is lost
        #
        #     Note: this is only possible with Solidity >= 0.4.12
        if is_compact_ast:
            expressions = [
                parse_expression(e, caller_context) if e else None for e in expression["components"]
            ]
        else:
            if "children" not in expression:
                attributes = expression["attributes"]
                components = attributes["components"]
                expressions = [
                    parse_expression(c, caller_context) if c else None for c in components
                ]
            else:
                expressions = [parse_expression(e, caller_context) for e in expression["children"]]
        # Add none for empty tuple items
        if "attributes" in expression:
            if "type" in expression["attributes"]:
                t = expression["attributes"]["type"]
                if ",," in t or "(," in t or ",)" in t:
                    t = t[len("tuple(") : -1]
                    elems = t.split(",")
                    for idx, _ in enumerate(elems):
                        if elems[idx] == "":
                            expressions.insert(idx, None)
        t = TupleExpression(expressions)
        t.set_offset(src, caller_context.compilation_unit)
        return t

    if name == "Conditional":
        if is_compact_ast:
            if_expression = parse_expression(expression["condition"], caller_context)
            then_expression = parse_expression(expression["trueExpression"], caller_context)
            else_expression = parse_expression(expression["falseExpression"], caller_context)
        else:
            children = expression["children"]
            assert len(children) == 3
            if_expression = parse_expression(children[0], caller_context)
            then_expression = parse_expression(children[1], caller_context)
            else_expression = parse_expression(children[2], caller_context)
        conditional = ConditionalExpression(if_expression, then_expression, else_expression)
        conditional.set_offset(src, caller_context.compilation_unit)
        return conditional

    if name == "Assignment":
        if is_compact_ast:
            left_expression = parse_expression(expression["leftHandSide"], caller_context)
            right_expression = parse_expression(expression["rightHandSide"], caller_context)

            operation_type = AssignmentOperationType.get_type(expression["operator"])

            operation_return_type = expression["typeDescriptions"]["typeString"]
        else:
            attributes = expression["attributes"]
            children = expression["children"]
            assert len(expression["children"]) == 2
            left_expression = parse_expression(children[0], caller_context)
            right_expression = parse_expression(children[1], caller_context)

            operation_type = AssignmentOperationType.get_type(attributes["operator"])
            operation_return_type = attributes["type"]

        assignement = AssignmentOperation(
            left_expression, right_expression, operation_type, operation_return_type
        )
        assignement.set_offset(src, caller_context.compilation_unit)
        return assignement

    if name == "Literal":
        subdenomination = None

        assert "children" not in expression

        if is_compact_ast:
            value = expression.get("value", None)
            if value:
                if expression.get("subdenomination"):
                    subdenomination = expression["subdenomination"]
            elif not value and value != "":
                value = "0x" + expression["hexValue"]
            type_candidate = expression["typeDescriptions"]["typeString"]

            # Length declaration for array was None until solc 0.5.5
            if type_candidate is None:
                if expression["kind"] == "number":
                    type_candidate = "int_const"
        else:
            value = expression["attributes"].get("value", None)
            if value:
                if (
                    "subdenomination" in expression["attributes"]
                    and expression["attributes"]["subdenomination"]
                ):
                    subdenomination = expression["attributes"]["subdenomination"]
            elif value is None:
                # for literal declared as hex
                # see https://solidity.readthedocs.io/en/v0.4.25/types.html?highlight=hex#hexadecimal-literals
                assert "hexvalue" in expression["attributes"]
                value = "0x" + expression["attributes"]["hexvalue"]
            type_candidate = expression["attributes"]["type"]

        if type_candidate is None:
            if value.isdecimal():
                type_candidate = ElementaryType("uint256")
            else:
                type_candidate = ElementaryType("string")
        elif type_candidate.startswith("rational_const "):
            type_candidate = ElementaryType("uint256")
        elif type_candidate.startswith("int_const "):
            type_candidate = ElementaryType("uint256")
        elif type_candidate.startswith("bool"):
            type_candidate = ElementaryType("bool")
        elif type_candidate.startswith("address"):
            type_candidate = ElementaryType("address")
        else:
            type_candidate = ElementaryType("string")
        literal = Literal(value, type_candidate, subdenomination)
        literal.set_offset(src, caller_context.compilation_unit)
        return literal

    if name == "Identifier":
        assert "children" not in expression

        t = None

        referenced_declaration = None
        if caller_context.is_compact_ast:
            value = expression["name"]
            t = expression["typeDescriptions"]["typeString"]
            if "referencedDeclaration" in expression:
                referenced_declaration = expression["referencedDeclaration"]
        else:
            value = expression["attributes"]["value"]
            if "type" in expression["attributes"]:
                t = expression["attributes"]["type"]
            if "referencedDeclaration" in expression["attributes"]:
                referenced_declaration = expression["attributes"]["referencedDeclaration"]

        if t:
            found = re.findall(
                r"(?:struct|enum|function|modifier) \(([\[\] ()a-zA-Z0-9\.,_]*)\)", t
            )
            assert len(found) <= 1
            if found:
                value = value + "(" + found[0] + ")"
                value = filter_name(value)

        var, was_created = find_variable(value, caller_context, referenced_declaration)
        if was_created:
            var.set_offset(src, caller_context.compilation_unit)

        identifier = Identifier(var)
        identifier.set_offset(src, caller_context.compilation_unit)
        var.references.append(identifier.source_mapping)

        return identifier

    if name == "IndexAccess":
        if is_compact_ast:
            # We dont use the index type here, as we recover it later
            # We could change the paradigm with the current AST parsing
            # And do the type parsing in advanced for most of the operation
            # index_type = expression["typeDescriptions"]["typeString"]
            left = expression["baseExpression"]
            right = expression.get("indexExpression", None)
        else:
            # index_type = expression["attributes"]["type"]
            children = expression["children"]
            left = children[0]
            right = children[1] if len(children) > 1 else None
        # IndexAccess is used to describe ElementaryTypeNameExpression
        # if abi.decode is used
        # For example, abi.decode(data, ...(uint[]) )
        if right is None:
            ret = parse_expression(left, caller_context)
            # Nested array are not yet available in abi.decode
            if isinstance(ret, ElementaryTypeNameExpression):
                old_type = ret.type
                ret.type = ArrayType(old_type, None)
            return ret

        left_expression = parse_expression(left, caller_context)
        right_expression = parse_expression(right, caller_context)
        index = IndexAccess(left_expression, right_expression)
        index.set_offset(src, caller_context.compilation_unit)
        return index

    if name == "MemberAccess":
        if caller_context.is_compact_ast:
            member_name = expression["memberName"]
            member_type = expression["typeDescriptions"]["typeString"]
            # member_type = parse_type(
            #     UnknownType(expression["typeDescriptions"]["typeString"]), caller_context
            # )
            member_expression = parse_expression(expression["expression"], caller_context)
        else:
            member_name = expression["attributes"]["member_name"]
            member_type = expression["attributes"]["type"]
            # member_type = parse_type(UnknownType(expression["attributes"]["type"]), caller_context)
            children = expression["children"]
            assert len(children) == 1
            member_expression = parse_expression(children[0], caller_context)
        if str(member_expression) == "super":
            super_name = parse_super_name(expression, is_compact_ast)
            var, was_created = find_variable(super_name, caller_context, is_super=True)
            if var is None:
                raise VariableNotFound(f"Variable not found: {super_name}")
            if was_created:
                var.set_offset(src, caller_context.compilation_unit)
            sup = SuperIdentifier(var)
            sup.set_offset(src, caller_context.compilation_unit)

            var.references.append(sup.source_mapping)

            return sup
        member_access = MemberAccess(member_name, member_type, member_expression)
        member_access.set_offset(src, caller_context.compilation_unit)
        # Only convert to SolidityVariableComposed if the base expression is NOT
        # a user-defined variable. This prevents a mistake like a parameter struct
        # named "self" in a function with a balance field being treated as Vyper's self.balance.
        is_user_defined = isinstance(member_expression, Identifier) and isinstance(
            member_expression.value,
            Variable,
        )
        if str(member_access) in SOLIDITY_VARIABLES_COMPOSED and not is_user_defined:
            id_idx = Identifier(SolidityVariableComposed(str(member_access)))
            id_idx.set_offset(src, caller_context.compilation_unit)
            return id_idx
        return member_access

    if name == "ElementaryTypeNameExpression":
        return _parse_elementary_type_name_expression(expression, is_compact_ast, caller_context)

    # NewExpression is not a root expression, it's always the child of another expression
    if name == "NewExpression":
        if is_compact_ast:
            type_name = expression["typeName"]
        else:
            children = expression["children"]
            assert len(children) == 1
            type_name = children[0]

        if type_name[caller_context.get_key()] == "ArrayTypeName":
            array_type = parse_type(type_name, caller_context)
            assert isinstance(array_type, ArrayType)
            array = NewArray(array_type)
            array.set_offset(src, caller_context.compilation_unit)
            return array

        if type_name[caller_context.get_key()] == "ElementaryTypeName":
            if is_compact_ast:
                elem_type = ElementaryType(type_name["name"])
            else:
                elem_type = ElementaryType(type_name["attributes"]["name"])
            new_elem = NewElementaryType(elem_type)
            new_elem.set_offset(src, caller_context.compilation_unit)
            return new_elem

        assert type_name[caller_context.get_key()] == "UserDefinedTypeName"

        contract_type = parse_type(type_name, caller_context)
        assert isinstance(contract_type, UserDefinedType)
        new = NewContract(contract_type)
        new.set_offset(src, caller_context.compilation_unit)
        return new

    if name == "ModifierInvocation":
        if is_compact_ast:
            called = parse_expression(expression["modifierName"], caller_context)
            arguments = []
            if expression.get("arguments", None):
                arguments = [parse_expression(a, caller_context) for a in expression["arguments"]]
        else:
            children = expression["children"]
            called = parse_expression(children[0], caller_context)
            arguments = [parse_expression(a, caller_context) for a in children[1::]]

        call = CallExpression(called, arguments, "Modifier")
        call.set_offset(src, caller_context.compilation_unit)
        return call

    if name == "IndexRangeAccess":
        # For now, we convert array slices to a direct array access
        # As a result the generated IR will lose the slices information
        # As far as I understand, array slice are only used in abi.decode
        # https://solidity.readthedocs.io/en/v0.6.12/types.html
        # TODO: Investigate array slices usage and implication for the IR
        base = parse_expression(expression["baseExpression"], caller_context)
        return base

    # Introduced with solc 0.8
    if name == "IdentifierPath":
        if caller_context.is_compact_ast:
            value = expression["name"]

            if "referencedDeclaration" in expression:
                referenced_declaration = expression["referencedDeclaration"]
            else:
                referenced_declaration = None

            var, was_created = find_variable(
                value, caller_context, referenced_declaration, is_identifier_path=True
            )
            if was_created:
                var.set_offset(src, caller_context.compilation_unit)

            identifier = Identifier(var)
            identifier.set_offset(src, caller_context.compilation_unit)

            var.references.append(identifier.source_mapping)

            return identifier

        raise ParsingError("IdentifierPath not currently supported for the legacy ast")

    raise ParsingError(f"Expression not parsed {name}")



from slither.exceptions import SlitherError
from slither.tools.upgradeability.checks.abstract_checks import (
    AbstractCheck,
    CheckClassification,
)
from slither.utils.function import get_function_id


def get_signatures(c):
    functions = c.functions
    functions = [
        f.full_name
        for f in functions
        if f.visibility in ["public", "external"] and not f.is_constructor and not f.is_fallback
    ]

    variables = c.state_variables
    variables = [
        variable.name + "()" for variable in variables if variable.visibility in ["public"]
    ]
    return list(set(functions + variables))


def _get_function_or_variable(contract, signature):
    f = contract.get_function_from_full_name(signature)

    if f:
        return f

    for variable in contract.state_variables:
        # Todo: can lead to incorrect variable in case of shadowing
        if variable.visibility in ["public"]:
            if variable.name + "()" == signature:
                return variable

    raise SlitherError(f"Function id checks: {signature} not found in {contract.name}")


class IDCollision(AbstractCheck):
    ARGUMENT = "function-id-collision"
    IMPACT = CheckClassification.HIGH

    HELP = "Functions ids collision"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#functions-ids-collisions"
    WIKI_TITLE = "Functions ids collisions"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect function id collision between the contract and the proxy.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    function gsf() public {
        // ...
    }
}

contract Proxy{
    function tgeo() public {
        // ...
    }
}
```
`Proxy.tgeo()` and `Contract.gsf()` have the same function id (0x67e43e43).
As a result, `Proxy.tgeo()` will shadow Contract.gsf()`.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Rename the function. Avoid public functions in the proxy.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_PROXY = True

    def _check(self):
        signatures_implem = get_signatures(self.contract)
        signatures_proxy = get_signatures(self.proxy)

        signatures_ids_implem = {get_function_id(s): s for s in signatures_implem}
        signatures_ids_proxy = {get_function_id(s): s for s in signatures_proxy}

        results = []

        for k, _ in signatures_ids_implem.items():
            if k in signatures_ids_proxy:
                if signatures_ids_implem[k] != signatures_ids_proxy[k]:
                    implem_function = _get_function_or_variable(
                        self.contract, signatures_ids_implem[k]
                    )
                    proxy_function = _get_function_or_variable(self.proxy, signatures_ids_proxy[k])

                    info = [
                        "Function id collision found: ",
                        implem_function,
                        " ",
                        proxy_function,
                        "\n",
                    ]
                    json = self.generate_result(info)
                    results.append(json)

        return results


class FunctionShadowing(AbstractCheck):
    ARGUMENT = "function-shadowing"
    IMPACT = CheckClassification.HIGH

    HELP = "Functions shadowing"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#functions-shadowing"
    WIKI_TITLE = "Functions shadowing"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect function shadowing between the contract and the proxy.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    function get() public {
        // ...
    }
}

contract Proxy{
    function get() public {
        // ...
    }
}
```
`Proxy.get` will shadow any call to `get()`. As a result `get()` is never executed in the logic contract and cannot be updated.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Rename the function. Avoid public functions in the proxy.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_PROXY = True

    def _check(self):
        signatures_implem = get_signatures(self.contract)
        signatures_proxy = get_signatures(self.proxy)

        signatures_ids_implem = {get_function_id(s): s for s in signatures_implem}
        signatures_ids_proxy = {get_function_id(s): s for s in signatures_proxy}

        results = []

        for k, _ in signatures_ids_implem.items():
            if k in signatures_ids_proxy:
                if signatures_ids_implem[k] == signatures_ids_proxy[k]:
                    implem_function = _get_function_or_variable(
                        self.contract, signatures_ids_implem[k]
                    )
                    proxy_function = _get_function_or_variable(self.proxy, signatures_ids_proxy[k])

                    info = [
                        "Function shadowing found: ",
                        implem_function,
                        " ",
                        proxy_function,
                        "\n",
                    ]
                    json = self.generate_result(info)
                    results.append(json)

        return results



# These are the default values in truffle
# https://www.trufflesuite.com/docs/truffle/getting-started/using-truffle-develop-and-the-console
OWNER_ADDRESS = "0x627306090abaB3A6e1400e9345bC60c78a8BEf57"
USER_ADDRESS = "0xf17f52151EbEF6C7334FAD080c5704D77216b732"
ATTACKER_ADDRESS = "0xC5fdf4076b8F3A5357c5E395ab970B5B54098Fef"


class Addresses:
    def __init__(
        self,
        owner: str | None = None,
        user: str | None = None,
        attacker: str | None = None,
    ):
        self.owner = owner if owner else OWNER_ADDRESS
        self.user = user if user else USER_ADDRESS
        self.attacker = attacker if attacker else ATTACKER_ADDRESS



import re

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.exceptions import FormatImpossible
from slither.formatters.utils.patches import create_patch

# Indicates the recommended versions for replacement
REPLACEMENT_VERSIONS = ["^0.4.25", "^0.5.3"]


# group:
# 0: ^ > >= < <= (optional)
# 1: ' ' (optional)
# 2: version number
# 3: version number
# 4: version number
PATTERN = re.compile(r"(\^|>|>=|<|<=)?([ ]+)?(\d+)\.(\d+)\.(\d+)")


def custom_format(slither: SlitherCompilationUnit, result: dict) -> None:
    elements = result["elements"]
    versions_used: list[str] = []
    for element in elements:
        versions_used.append("".join(element["type_specific_fields"]["directive"][1:]))
    solc_version_replace = _analyse_versions(versions_used)
    for element in elements:
        _patch(
            slither,
            result,
            element["source_mapping"]["filename_absolute"],
            solc_version_replace,
            element["source_mapping"]["start"],
            element["source_mapping"]["start"] + element["source_mapping"]["length"],
        )


def _analyse_versions(used_solc_versions: list[str]) -> str:
    replace_solc_versions = []
    for version in used_solc_versions:
        replace_solc_versions.append(_determine_solc_version_replacement(version))
    if not all(version == replace_solc_versions[0] for version in replace_solc_versions):
        raise FormatImpossible("Multiple incompatible versions!")
    return replace_solc_versions[0]


def _determine_solc_version_replacement(used_solc_version: str) -> str:
    versions = PATTERN.findall(used_solc_version)
    if len(versions) == 1:
        version = versions[0]
        minor_version = ".".join(version[2:])[2]
        if minor_version == "4":
            return "pragma solidity " + REPLACEMENT_VERSIONS[0] + ";"
        if minor_version == "5":
            return "pragma solidity " + REPLACEMENT_VERSIONS[1] + ";"
        raise FormatImpossible("Unknown version!")
    if len(versions) == 2:
        version_right = versions[1]
        minor_version_right = ".".join(version_right[2:])[2]
        if minor_version_right == "4":
            # Replace with 0.4.25
            return "pragma solidity " + REPLACEMENT_VERSIONS[0] + ";"
        if minor_version_right in ["5", "6"]:
            # Replace with 0.5.3
            return "pragma solidity " + REPLACEMENT_VERSIONS[1] + ";"
    raise FormatImpossible("Unknown version!")


def _patch(
    slither: SlitherCompilationUnit,
    result: dict,
    in_file: str,
    pragma: str | bytes,
    modify_loc_start: int,
    modify_loc_end: int,
) -> None:
    in_file_str = slither.core.source_code[in_file].encode("utf8")
    old_str_of_interest = in_file_str[modify_loc_start:modify_loc_end]
    create_patch(
        result,
        in_file,
        int(modify_loc_start),
        int(modify_loc_end),
        old_str_of_interest,
        pragma,
    )



from slither.core.declarations import SolidityFunction, Function
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import SolidityCall
from slither.utils.output import Output


def _assembly_node(function: Function) -> SolidityCall | None:
    """
    Check if there is a node that use return in assembly

    Args:
        function:

    Returns:

    """

    for ir in function.all_solidity_calls():
        if ir.function == SolidityFunction("return(uint256,uint256)"):
            return ir
    return None


class IncorrectReturn(AbstractDetector):
    """
    Check for cases where a return(a,b) is used in an assembly function
    """

    ARGUMENT = "incorrect-return"
    HELP = "If a `return` is incorrectly used in assembly mode."
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-return-in-assembly"
    )

    WIKI_TITLE = "Incorrect return in assembly"
    WIKI_DESCRIPTION = "Detect if `return` in an assembly block halts unexpectedly the execution."
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {
    function f() internal returns (uint a, uint b) {
        assembly {
            return (5, 6)
        }
    }

    function g() returns (bool){
        f();
        return true;
    }
}
```
The return statement in `f` will cause execution in `g` to halt.
The function will return 6 bytes starting from offset 5, instead of returning a boolean."""

    WIKI_RECOMMENDATION = "Use the `leave` statement."

    def _detect(self) -> list[Output]:
        results: list[Output] = []
        for c in self.contracts:
            for f in c.functions_and_modifiers_declared:
                for ir in f.internal_calls:
                    if ir.node.sons:
                        function_called = ir.function
                        if isinstance(function_called, Function):
                            found = _assembly_node(function_called)
                            if found:
                                info: DETECTOR_INFO = [
                                    f,
                                    " calls ",
                                    function_called,
                                    " which halt the execution ",
                                    found.node,
                                    "\n",
                                ]
                                json = self.generate_result(info)

                                results.append(json)

        return results



from importlib import metadata
import json
from typing import Any
import urllib

from packaging.version import parse, Version

from slither.utils.colors import yellow, green


def get_installed_version(name: str) -> Version | None:
    try:
        return parse(metadata.version(name))
    except metadata.PackageNotFoundError:
        return None


def get_github_version(name: str) -> Version | None:
    try:
        # type: ignore
        with urllib.request.urlopen(
            f"https://api.github.com/repos/crytic/{name}/releases/latest"
        ) as response:
            text = response.read()
            data = json.loads(text)
            return parse(data["tag_name"])
    except Exception:
        return None


def show_versions(**_kwargs: Any) -> None:
    versions = {
        "Slither": (get_installed_version("slither-analyzer"), get_github_version("slither")),
        "crytic-compile": (
            get_installed_version("crytic-compile"),
            get_github_version("crytic-compile"),
        ),
        "solc-select": (get_installed_version("solc-select"), get_github_version("solc-select")),
    }

    outdated = {
        name
        for name, (installed, latest) in versions.items()
        if not installed or not latest or latest > installed
    }

    for name, (installed, latest) in versions.items():
        color = yellow if name in outdated else green
        print(
            f"{name + ':':<16}{color(str(installed) or 'N/A'):<16} (latest is {str(latest) or 'Unknown'})"
        )

    if len(outdated) > 0:
        print()
        print(
            yellow(
                f"Please update {', '.join(outdated)} to the latest release before creating a bug report."
            )
        )
    else:
        print()
        print(green("Your tools are up to date."))



from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.event_variable import EventVariable


class Event(SourceMapping):
    def __init__(self) -> None:
        super().__init__()
        self._name = None
        self._elems: list[EventVariable] = []

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def signature(self) -> tuple[str, list[str]]:
        """Return the function signature
        Returns:
            (str, list(str)): name, list parameters type
        """
        return self.name, [str(x.type) for x in self.elems]

    @property
    def full_name(self) -> str:
        """Return the function signature as a str
        Returns:
            str: func_name(type1,type2)
        """
        name, parameters = self.signature
        return name + "(" + ",".join(parameters) + ")"

    @property
    def elems(self) -> list["EventVariable"]:
        return self._elems

    def __str__(self) -> str:
        return self.name



import argparse
import inspect
import logging
import os
import shutil
import sys
import time
from pathlib import Path
from typing import Any
from crytic_compile import cryticparser
from slither import Slither
from slither.tools.mutator.utils.testing_generated_mutant import run_test_cmd
from slither.tools.mutator.mutators import all_mutators
from slither.utils.colors import blue, green, magenta, red
from .mutators.abstract_mutator import AbstractMutator
from .utils.command_line import output_mutators
from .utils.file_handling import (
    transfer_and_delete,
    backup_source_file,
    get_sol_file_list,
)

logging.basicConfig()
logger = logging.getLogger("Slither-Mutate")
logger.setLevel(logging.INFO)

###################################################################################
###################################################################################
# region Cli Arguments
###################################################################################
###################################################################################


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    Returns: The arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Experimental smart contract mutator. Based on https://arxiv.org/abs/2006.11597",
        usage="slither-mutate <codebase> --test-cmd <test command> <options>",
    )

    parser.add_argument("codebase", help="Codebase to analyze (.sol file, project directory, ...)")

    parser.add_argument(
        "--list-mutators",
        help="List available detectors",
        action=ListMutators,
        nargs=0,
        default=False,
    )

    # argument to add the test command
    parser.add_argument("--test-cmd", help="Command to run the tests for your project")

    # argument to add the test directory - containing all the tests
    parser.add_argument("--test-dir", help="Tests directory")

    # argument to ignore the interfaces, libraries
    parser.add_argument("--ignore-dirs", help="Directories to ignore")

    # time out argument
    parser.add_argument("--timeout", help="Set timeout for test command (by default 30 seconds)")

    # output directory argument
    parser.add_argument(
        "--output-dir", help="Name of output directory (by default 'mutation_campaign')"
    )

    # to print just all the mutants
    parser.add_argument(
        "-v",
        "--verbose",
        help="log mutants that are caught, uncaught, and fail to compile",
        action="store_true",
        default=False,
    )

    # select list of mutators to run
    parser.add_argument(
        "--mutators-to-run",
        help="mutant generators to run",
    )

    # list of contract names you want to mutate
    parser.add_argument(
        "--contract-names",
        help="list of contract names you want to mutate",
    )

    # target specific functions by selector
    parser.add_argument(
        "--target-functions",
        help="Comma-separated list of function selectors (hex like 0xa9059cbb or signature like transfer(address,uint256))",
    )

    # flag to run full mutation based revert mutator output
    parser.add_argument(
        "--comprehensive",
        help="continue testing minor mutations if severe mutants are uncaught",
        action="store_true",
        default=False,
    )

    # Initiate all the crytic config cli options
    cryticparser.init(parser)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def _get_mutators(mutators_list: list[str] | None) -> list[type[AbstractMutator]]:
    detectors_ = [getattr(all_mutators, name) for name in dir(all_mutators)]
    if mutators_list is not None:
        detectors = [
            c
            for c in detectors_
            if inspect.isclass(c)
            and issubclass(c, AbstractMutator)
            and str(c.NAME) in mutators_list
        ]
    else:
        detectors = [c for c in detectors_ if inspect.isclass(c) and issubclass(c, AbstractMutator)]
    return detectors


class ListMutators(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        checks = _get_mutators(None)
        output_mutators(checks)
        parser.exit()


# endregion
###################################################################################
###################################################################################
# region Selector Parsing
###################################################################################
###################################################################################


def parse_target_selectors(selector_str: str) -> set[int]:
    """Parse comma-separated selectors (hex or signature format)

    Handles signatures with commas like transfer(address,uint256) by
    only splitting on commas outside of parentheses.
    """
    from slither.utils.function import get_function_id

    selectors: set[int] = set()

    # Split on commas only when not inside parentheses
    parts = []
    current = ""
    depth = 0
    for char in selector_str:
        if char == "(":
            depth += 1
            current += char
        elif char == ")":
            depth -= 1
            current += char
        elif char == "," and depth == 0:
            parts.append(current.strip())
            current = ""
        else:
            current += char
    if current.strip():
        parts.append(current.strip())

    for s in parts:
        if not s:
            continue
        if s.startswith("0x"):
            # Hex format: 0xa9059cbb
            if len(s) != 10:
                logger.error(f"Invalid selector format: {s} (must be 0x + 8 hex chars)")
                sys.exit(1)
            try:
                selectors.add(int(s, 16))
            except ValueError:
                logger.error(f"Invalid hex selector: {s}")
                sys.exit(1)
        else:
            # Signature format: transfer(address,uint256)
            selectors.add(get_function_id(s))
    return selectors


# endregion
###################################################################################
###################################################################################
# region Main
###################################################################################
###################################################################################


def main() -> None:
    args = parse_args()

    # arguments
    test_command: str = args.test_cmd
    test_directory: str | None = args.test_dir
    paths_to_ignore: str | None = args.ignore_dirs
    output_dir: str | None = args.output_dir
    timeout: int | None = args.timeout
    solc_remappings: str | None = args.solc_remaps
    verbose: bool | None = args.verbose
    mutators_to_run: list[str] | None = args.mutators_to_run
    comprehensive_flag: bool | None = args.comprehensive

    logger.info(blue(f"Starting mutation campaign in {args.codebase}"))

    if paths_to_ignore:
        paths_to_ignore_list = paths_to_ignore.strip("][").split(",")
    else:
        paths_to_ignore_list = []

    contract_names: list[str] = []
    if args.contract_names:
        contract_names = args.contract_names.split(",")

    target_selectors: set[int] | None = None
    if args.target_functions:
        target_selectors = parse_target_selectors(args.target_functions)

    # get all the contracts as a list from given codebase
    sol_file_list: list[str] = get_sol_file_list(Path(args.codebase), paths_to_ignore_list)

    if not contract_names:
        logger.info(blue("Preparing to mutate files:\n- " + "\n- ".join(sol_file_list)))
    else:
        logger.info(blue("Preparing to mutate contracts:\n- " + "\n- ".join(contract_names)))

    # folder where backup files and uncaught mutants are saved
    if output_dir is None:
        output_dir = "./mutation_campaign"

    output_folder = Path(output_dir).resolve()
    if output_folder.is_dir():
        shutil.rmtree(output_folder)

    # setting RR mutator as first mutator
    mutators_list = _get_mutators(mutators_to_run)

    # insert RR and CR in front of the list
    CR_RR_list = []
    duplicate_list = mutators_list.copy()
    for M in duplicate_list:
        if M.NAME == "RR":
            mutators_list.remove(M)
            CR_RR_list.insert(0, M)
        elif M.NAME == "CR":
            mutators_list.remove(M)
            CR_RR_list.insert(1, M)
    mutators_list = CR_RR_list + mutators_list

    logger.info(blue("Timing tests.."))

    # run and time tests, abort if they're broken
    start_time = time.time()
    # no timeout or target_file during the first run, but be verbose on failure
    if not run_test_cmd(test_command, None, None, True):
        logger.error(red("Test suite fails with mutations, aborting"))
        return
    elapsed_time = round(time.time() - start_time)

    # set default timeout
    # default to twice as long as it usually takes to run the test suite
    if timeout is None:
        timeout = int(elapsed_time * 2)
    else:
        timeout = int(timeout)
        if timeout < elapsed_time:
            logger.info(
                red(
                    f"Provided timeout {timeout} is too short for tests that run in {elapsed_time} seconds"
                )
            )
            return

    logger.info(
        green(
            f"Test suite passes in {elapsed_time} seconds, commencing mutation campaign with a timeout of {timeout} seconds\n"
        )
    )

    # Keep a list of all already mutated contracts so we don't mutate them twice
    mutated_contracts: list[str] = []

    for filename in sol_file_list:
        file_name = os.path.split(filename)[1].split(".sol")[0]
        # slither object
        sl = Slither(filename, **vars(args))
        # create a backup files
        files_dict = backup_source_file(sl.source_code, output_folder)
        # total revert/comment/tweak mutants that were generated and compiled
        total_mutant_counts = [0, 0, 0]
        # total caught revert/comment/tweak mutants
        caught_mutant_counts = [0, 0, 0]
        # lines those need not be mutated (taken from RR and CR)
        dont_mutate_lines = []

        # perform mutations on {target_contract} in file {file_name}
        # setup placeholder val to signal whether we need to skip if no target_contract is found
        skip_flag = "SLITHER_SKIP_MUTATIONS"
        target_contract = skip_flag if contract_names else ""
        try:
            # loop through all contracts in file_name
            for compilation_unit_of_main_file in sl.compilation_units:
                for contract in compilation_unit_of_main_file.contracts:
                    if contract.name in contract_names and contract.name not in mutated_contracts:
                        target_contract = contract
                        break
                    if not contract_names and contract.name.lower() == file_name.lower():
                        target_contract = contract
                        break

                if target_contract == "":
                    logger.info(
                        f"Cannot find contracts in file {filename}, try specifying them with --contract-names"
                    )
                    continue

                if target_contract == skip_flag:
                    continue

                if target_contract.is_interface:
                    logger.debug(f"Skipping mutations on interface {filename}")
                    continue

                # Add our target to the mutation list
                mutated_contracts.append(target_contract.name)
                logger.info(blue(f"Mutating contract {target_contract}"))

                # Validate target selectors and collect target modifiers
                target_modifiers: set[str] | None = None
                if target_selectors:
                    from slither.utils.function import get_function_id

                    matching_functions = []
                    target_modifiers = set()

                    for func in target_contract.functions_declared:
                        func_selector = get_function_id(func.solidity_signature)
                        if func_selector in target_selectors:
                            matching_functions.append(func)
                            # Collect modifiers used by this function
                            for mod in func.modifiers:
                                target_modifiers.add(mod.name)

                    if not matching_functions:
                        logger.error(
                            f"No functions in {target_contract.name} match selectors: {[hex(s) for s in target_selectors]}"
                        )
                        sys.exit(1)

                    logger.info(
                        blue(f"Targeting functions: {[f.name for f in matching_functions]}")
                    )
                    if target_modifiers:
                        logger.info(blue(f"Including modifiers: {list(target_modifiers)}"))

                for M in mutators_list:
                    m = M(
                        compilation_unit_of_main_file,
                        int(timeout),
                        test_command,
                        test_directory,
                        target_contract,
                        solc_remappings,
                        verbose,
                        output_folder,
                        dont_mutate_lines,
                        target_selectors=target_selectors,
                        target_modifiers=target_modifiers,
                    )
                    (total_counts, caught_counts, lines_list) = m.mutate()

                    if m.NAME == "RR":
                        total_mutant_counts[0] += total_counts[0]
                        caught_mutant_counts[0] += caught_counts[0]
                        if verbose:
                            logger.info(
                                magenta(
                                    f"Mutator {m.NAME} found {caught_counts[0]} caught revert mutants (out of {total_counts[0]} that compile)"
                                )
                            )
                    elif m.NAME == "CR":
                        total_mutant_counts[1] += total_counts[1]
                        caught_mutant_counts[1] += caught_counts[1]
                        if verbose:
                            logger.info(
                                magenta(
                                    f"Mutator {m.NAME} found {caught_counts[1]} caught comment mutants (out of {total_counts[1]} that compile)"
                                )
                            )
                    else:
                        total_mutant_counts[2] += total_counts[2]
                        caught_mutant_counts[2] += caught_counts[2]
                        if verbose:
                            logger.info(
                                magenta(
                                    f"Mutator {m.NAME} found {caught_counts[2]} caught tweak mutants (out of {total_counts[2]} that compile)"
                                )
                            )
                            logger.info(
                                magenta(
                                    f"Running total: found {caught_mutant_counts[2]} caught tweak mutants (out of {total_mutant_counts[2]} that compile)"
                                )
                            )

                    dont_mutate_lines = lines_list
                    if comprehensive_flag:
                        dont_mutate_lines = []

        except Exception as e:
            logger.error(e)
            transfer_and_delete(files_dict)

        except KeyboardInterrupt:
            # transfer and delete the backup files if interrupted
            logger.error("\nExecution interrupted by user (Ctrl + C). Cleaning up...")
            transfer_and_delete(files_dict)

        # transfer and delete the backup files
        transfer_and_delete(files_dict)

        if target_contract == skip_flag:
            logger.debug(f"No target contracts found in {filename}, skipping")
            continue

        # log results for this file
        logger.info(blue(f"Done mutating {target_contract}."))
        if total_mutant_counts[0] > 0:
            logger.info(
                magenta(
                    f"Revert mutants: {caught_mutant_counts[0]} caught of {total_mutant_counts[0]} ({round(100 * caught_mutant_counts[0] / total_mutant_counts[0], 1)}% caught)"
                )
            )
        else:
            logger.info(magenta("Zero Revert mutants analyzed"))

        if total_mutant_counts[1] > 0:
            logger.info(
                magenta(
                    f"Comment mutants: {caught_mutant_counts[1]} caught of {total_mutant_counts[1]} ({round(100 * caught_mutant_counts[1] / total_mutant_counts[1], 1)}% caught)"
                )
            )
        else:
            logger.info(magenta("Zero Comment mutants analyzed"))

        if total_mutant_counts[2] > 0:
            logger.info(
                magenta(
                    f"Tweak mutants: {caught_mutant_counts[2]} caught of {total_mutant_counts[2]} ({round(100 * caught_mutant_counts[2] / total_mutant_counts[2], 1)}% caught)"
                )
            )
        else:
            logger.info(magenta("Zero Tweak mutants analyzed\n"))

        # Reset mutant counts before moving on to the next file
        total_mutant_counts[0] = 0
        total_mutant_counts[1] = 0
        total_mutant_counts[2] = 0
        caught_mutant_counts[0] = 0
        caught_mutant_counts[1] = 0
        caught_mutant_counts[2] = 0

    # Print the total time elapsed in a human-readable time format
    elapsed_time = round(time.time() - start_time)
    hours, remainder = divmod(elapsed_time, 3600)
    minutes, seconds = divmod(remainder, 60)
    if hours > 0:
        elapsed_string = f"{hours} {'hour' if hours == 1 else 'hours'}"
    elif minutes > 0:
        elapsed_string = f"{minutes} {'minute' if minutes == 1 else 'minutes'}"
    else:
        elapsed_string = f"{seconds} {'second' if seconds == 1 else 'seconds'}"

    logger.info(
        blue(f"Finished mutation testing assessment of '{args.codebase}' in {elapsed_string}\n")
    )


# endregion



"""
Module detecting uninitialized storage variables

Recursively explore the CFG to only report uninitialized storage variables that are
written before being read
"""

from slither.core.cfg.node import Node
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.utils.output import Output


class UninitializedStorageVars(AbstractDetector):
    ARGUMENT = "uninitialized-storage"
    HELP = "Uninitialized storage variables"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#uninitialized-storage-variables"

    WIKI_TITLE = "Uninitialized storage variables"
    WIKI_DESCRIPTION = "An uninitialized storage variable will act as a reference to the first state variable, and can override a critical variable."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Uninitialized{
    address owner = msg.sender;

    struct St{
        uint a;
    }

    function func() {
        St st;
        st.a = 0x0;
    }
}
```
Bob calls `func`. As a result, `owner` is overridden to `0`.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Initialize all storage variables."

    # node.context[self.key] contains the uninitialized storage variables
    key = "UNINITIALIZEDSTORAGE"

    def _detect_uninitialized(
        self, function: FunctionContract, node: Node, visited: list[Node]
    ) -> None:
        if node in visited:
            return

        visited = visited + [node]

        fathers_context = []

        for father in node.fathers:
            if self.key in father.context:
                fathers_context += father.context[self.key]

        # Exclude paths that dont bring further information
        if node in self.visited_all_paths:
            if all(f_c in self.visited_all_paths[node] for f_c in fathers_context):
                return
        else:
            self.visited_all_paths[node] = []

        self.visited_all_paths[node] = list(set(self.visited_all_paths[node] + fathers_context))

        if self.key in node.context:
            fathers_context += node.context[self.key]

        variables_read = node.variables_read
        for uninitialized_storage_variable in fathers_context:
            if uninitialized_storage_variable in variables_read:
                self.results.append((function, uninitialized_storage_variable))

        # Only save the storage variables that are not yet written
        uninitialized_storage_variables = list(set(fathers_context) - set(node.variables_written))
        node.context[self.key] = uninitialized_storage_variables

        for son in node.sons:
            self._detect_uninitialized(function, son, visited)

    def _detect(self) -> list[Output]:
        """Detect uninitialized storage variables

        Recursively visit the calls
        Returns:
            dict: [contract name] = set(storage variable uninitialized)
        """
        results = []

        self.results = []
        self.visited_all_paths = {}

        for contract in self.compilation_unit.contracts:
            for function in contract.functions:
                if function.is_implemented and function.entry_point:
                    locals_except_params = set(function.variables) - set(function.parameters)
                    uninitialized_storage_variables = [
                        v for v in locals_except_params if v.is_storage and v.uninitialized
                    ]

                    function.entry_point.context[self.key] = uninitialized_storage_variables
                    self._detect_uninitialized(function, function.entry_point, [])

        for function, uninitialized_storage_variable in self.results:
            info = [
                uninitialized_storage_variable,
                " is a storage variable never initialized\n",
            ]
            json = self.generate_result(info)
            results.append(json)

        return results



"""
Special variable to model import with renaming
"""

from slither.core.declarations import Import
from slither.core.declarations.contract import Contract
from slither.core.declarations.solidity_variables import SolidityVariable
from slither.core.solidity_types import ElementaryType
from slither.core.variables.variable import Variable


class SolidityImportPlaceHolder(Variable):
    """
    Placeholder for import on top level objects
    See the example at https://blog.soliditylang.org/2020/09/02/solidity-0.7.1-release-announcement/
    In the long term we should remove this and better integrate import aliases
    """

    def __init__(self, import_directive: Import) -> None:
        super().__init__()
        assert import_directive.alias is not None
        self._import_directive = import_directive
        self._name = import_directive.alias
        self._type = ElementaryType("string")
        self._initialized = True
        self._visibility = "private"
        self._is_constant = True

    @property
    def type(self) -> ElementaryType:
        return ElementaryType("string")

    def __eq__(self, other: Contract | SolidityVariable) -> bool:
        return (
            self.__class__ == other.__class__
            and self._import_directive.filename == self._import_directive.filename
        )

    @property
    def import_directive(self) -> Import:
        return self._import_directive

    def __hash__(self):
        return hash(str(self.import_directive))



"""
Module detecting numbers with too many digits.
"""

import re

from slither.core.cfg.node import Node
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.variables import Constant
from slither.utils.output import Output

_HEX_ADDRESS_REGEXP = re.compile("(0[xX])?[0-9a-fA-F]{40}")


def is_hex_address(value: str) -> bool:
    """
    Checks if the given string of text type is an address in hexadecimal encoded form.
    """
    return _HEX_ADDRESS_REGEXP.fullmatch(value) is not None


class TooManyDigits(AbstractDetector):
    """
    Detect numbers with too many digits
    """

    ARGUMENT = "too-many-digits"
    HELP = "Conformance to numeric notation best practices"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#too-many-digits"
    WIKI_TITLE = "Too many digits"

    # region wiki_description
    WIKI_DESCRIPTION = """
Literals with many digits are difficult to read and review. Use scientific notation or suffixes to make the code more readable.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract MyContract{
    uint 1_ether = 10000000000000000000;
}
```

While `1_ether` looks like `1 ether`, it is `10 ether`. As a result, it's likely to be used incorrectly.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Use:
- [Ether suffix](https://solidity.readthedocs.io/en/latest/units-and-global-variables.html#ether-units),
- [Time suffix](https://solidity.readthedocs.io/en/latest/units-and-global-variables.html#time-units), or
- [The scientific notation](https://solidity.readthedocs.io/en/latest/types.html#rational-and-integer-literals)
"""
    # endregion wiki_recommendation

    @staticmethod
    def _detect_too_many_digits(f: FunctionContract) -> list[Node]:
        ret = []
        for node in f.nodes:
            # each node contains a list of IR instruction
            for ir in node.irs:
                # iterate over all the variables read by the IR
                for read in ir.read:
                    # if the variable is a constant
                    if isinstance(read, Constant):
                        # read.value can return an int or a str. Convert it to str
                        value_as_str = read.original_value
                        if "00000" in value_as_str and not is_hex_address(value_as_str):
                            # Info to be printed
                            ret.append(node)
        return ret

    def _detect(self) -> list[Output]:
        results = []

        # iterate over all contracts
        for contract in self.compilation_unit.contracts_derived:
            # iterate over all functions
            for f in contract.functions:
                # iterate over all the nodes
                ret = self._detect_too_many_digits(f)
                if ret:
                    func_info: DETECTOR_INFO = [f, " uses literals with too many digits:"]
                    for node in ret:
                        node_info: DETECTOR_INFO = func_info + ["\n\t- ", node, "\n"]

                        # Add the result in result
                        res = self.generate_result(node_info)
                        results.append(res)

        return results



"""
Module detecting shadowing variables on abstract contract
Recursively check the called functions
"""

from slither.core.declarations import Contract
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.utils.output import Output, AllSupportedOutput
from .common import is_upgradable_gap_variable


def detect_shadowing(contract: Contract) -> list[list[StateVariable]]:
    ret: list[list[StateVariable]] = []
    variables_fathers = []
    for father in contract.inheritance:
        if all(not f.is_implemented for f in father.functions + list(father.modifiers)):
            variables_fathers += father.state_variables_declared

    var: StateVariable
    for var in contract.state_variables_declared:
        if is_upgradable_gap_variable(contract, var):
            continue

        shadow: list[StateVariable] = [v for v in variables_fathers if v.name == var.name]
        if shadow:
            ret.append([var] + shadow)
    return ret


class ShadowingAbstractDetection(AbstractDetector):
    """
    Shadowing detection
    """

    ARGUMENT = "shadowing-abstract"
    HELP = "State variables shadowing from abstract contracts"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#state-variable-shadowing-from-abstract-contracts"

    WIKI_TITLE = "State variable shadowing from abstract contracts"
    WIKI_DESCRIPTION = "Detection of state variables shadowed from abstract contracts."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract BaseContract{
    address owner;
}

contract DerivedContract is BaseContract{
    address owner;
}
```
`owner` of `BaseContract` is shadowed in `DerivedContract`."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove the state variable shadowing."

    def _detect(self) -> list[Output]:
        """Detect shadowing

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func', 'shadow'}

        """
        results: list[Output] = []
        for contract in self.contracts:
            shadowing = detect_shadowing(contract)
            if shadowing:
                for all_variables in shadowing:
                    shadow = all_variables[0]
                    variables = all_variables[1:]
                    info: list[AllSupportedOutput] = [shadow, " shadows:\n"]
                    for var in variables:
                        info += ["\t- ", var, "\n"]

                    res = self.generate_result(info)

                    results.append(res)

        return results



"""
Module printing summary of the contract
"""

import collections

from slither.core.declarations import FunctionContract
from slither.printers.abstract_printer import AbstractPrinter
from slither.utils import output
from slither.utils.colors import blue, green, magenta
from slither.utils.output import Output


class ContractSummary(AbstractPrinter):
    ARGUMENT = "contract-summary"
    HELP = "Print a summary of the contracts"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#contract-summary"

    def output(self, _filename: str) -> Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""

        all_contracts = []
        for c in self.contracts:
            is_upgradeable_proxy = c.is_upgradeable_proxy
            is_upgradeable = c.is_upgradeable

            additional_txt_info = ""

            if is_upgradeable_proxy:
                additional_txt_info += " (Upgradeable Proxy)"

            if is_upgradeable:
                additional_txt_info += " (Upgradeable)"

            if c in self.slither.contracts_derived:
                additional_txt_info += " (Most derived contract)"

            txt += blue(f"\n+ Contract {c.name}{additional_txt_info}\n")
            additional_fields = output.Output(
                "",
                additional_fields={
                    "is_upgradeable_proxy": is_upgradeable_proxy,
                    "is_upgradeable": is_upgradeable,
                    "is_most_derived": c in self.slither.contracts_derived,
                },
            )

            # Order the function with
            # contract_declarer -> list_functions
            public_function = [
                (f.contract_declarer.name, f)
                for f in c.functions
                if (not f.is_shadowed and not f.is_constructor_variables)
            ]
            collect: dict[str, list[FunctionContract]] = collections.defaultdict(list)
            for a, b in public_function:
                collect[a].append(b)

            for contract, functions in collect.items():
                txt += blue(f"  - From {contract}\n")

                functions = sorted(functions, key=lambda f: f.full_name)

                for function in functions:
                    if function.visibility in ["external", "public"]:
                        txt += green(f"    - {function.full_name} ({function.visibility})\n")
                    if function.visibility in ["internal", "private"]:
                        txt += magenta(f"    - {function.full_name} ({function.visibility})\n")
                    if function.visibility not in [
                        "external",
                        "public",
                        "internal",
                        "private",
                    ]:
                        txt += f"    - {function.full_name}  ({function.visibility})\n"

                    additional_fields.add(
                        function, additional_fields={"visibility": function.visibility}
                    )

            all_contracts.append((c, additional_fields.data))

        self.info(txt)

        res = self.generate_output(txt)
        for current_contract, current_additional_fields in all_contracts:
            res.add(current_contract, additional_fields=current_additional_fields)

        return res



from typing import Any

from slither.core.expressions.expression import Expression


class CallExpression(Expression):
    def __init__(
        self,
        called: Expression,
        arguments: list[Any],
        type_call: str,
        names: list[str] | None = None,
    ) -> None:
        """
        #### Parameters
        called -
            The expression denoting the function to be called
        arguments -
            List of argument expressions
        type_call -
            A string formatting of the called function's return type
        names -
            For calls with named fields, list fields in call order.
            For calls without named fields, None.
        """
        assert isinstance(called, Expression)
        assert (names is None) or isinstance(names, list)
        super().__init__()
        self._called: Expression = called
        self._arguments: list[Expression] = arguments
        self._type_call: str = type_call
        self._names: list[str] | None = names
        # gas and value are only available if the syntax is {gas: , value: }
        # For the .gas().value(), the member are considered as function call
        # And converted later to the correct info (convert.py)
        self._gas: Expression | None = None
        self._value: Expression | None = None
        self._salt: Expression | None = None

    @property
    def names(self) -> list[str] | None:
        """
        For calls with named fields, list fields in call order.
        For calls without named fields, None.
        """
        return self._names

    @property
    def call_value(self) -> Expression | None:
        return self._value

    @call_value.setter
    def call_value(self, v: Expression | None) -> None:
        self._value = v

    @property
    def call_gas(self) -> Expression | None:
        return self._gas

    @call_gas.setter
    def call_gas(self, gas: Expression | None) -> None:
        self._gas = gas

    @property
    def call_salt(self) -> Expression | None:
        return self._salt

    @call_salt.setter
    def call_salt(self, salt: Expression | None) -> None:
        self._salt = salt

    @property
    def called(self) -> Expression:
        return self._called

    @property
    def arguments(self) -> list[Expression]:
        return self._arguments

    @property
    def type_call(self) -> str:
        return self._type_call

    def __str__(self) -> str:
        txt = str(self._called)
        if self.call_gas or self.call_value:
            gas = f"gas: {self.call_gas}" if self.call_gas else ""
            value = f"value: {self.call_value}" if self.call_value else ""
            salt = f"salt: {self.call_salt}" if self.call_salt else ""
            if gas or value or salt:
                options = [gas, value, salt]
                txt += "{" + ",".join([o for o in options if o != ""]) + "}"
        args = (
            "{"
            + ",".join([f"{n}:{a!s}" for (a, n) in zip(self._arguments, self._names, strict=False)])
            + "}"
            if self._names is not None
            else ",".join([str(a) for a in self._arguments])
        )
        return txt + "(" + args + ")"



from slither.core.declarations.solidity_variables import SOLIDITY_FUNCTIONS
from slither.core.expressions import BinaryOperationType, UnaryOperationType

# taken from https://github.com/ethereum/solidity/blob/e11b9ed9f2c254bc894d844c0a64a0eb76bbb4fd/libevmasm/Instruction.cpp#L184
evm_opcodes = [
    "STOP",
    "ADD",
    "SUB",
    "MUL",
    "DIV",
    "SDIV",
    "MOD",
    "SMOD",
    "EXP",
    "NOT",
    "LT",
    "GT",
    "SLT",
    "SGT",
    "EQ",
    "ISZERO",
    "AND",
    "OR",
    "XOR",
    "BYTE",
    "SHL",
    "SHR",
    "SAR",
    "CLZ",
    "ADDMOD",
    "MULMOD",
    "SIGNEXTEND",
    "KECCAK256",
    "ADDRESS",
    "BALANCE",
    "ORIGIN",
    "CALLER",
    "CALLVALUE",
    "CALLDATALOAD",
    "CALLDATASIZE",
    "CALLDATACOPY",
    "CODESIZE",
    "CODECOPY",
    "GASPRICE",
    "EXTCODESIZE",
    "EXTCODECOPY",
    "RETURNDATASIZE",
    "RETURNDATACOPY",
    "MCOPY",
    "EXTCODEHASH",
    "BLOCKHASH",
    "COINBASE",
    "TIMESTAMP",
    "NUMBER",
    "DIFFICULTY",
    "PREVRANDAO",
    "GASLIMIT",
    "CHAINID",
    "SELFBALANCE",
    "BASEFEE",
    "BLOBHASH",
    "BLOBBASEFEE",
    "POP",
    "MLOAD",
    "MSTORE",
    "MSTORE8",
    "SLOAD",
    "SSTORE",
    "TLOAD",
    "TSTORE",
    "JUMP",
    "JUMPI",
    "PC",
    "MSIZE",
    "GAS",
    "JUMPDEST",
    "PUSH1",
    "PUSH2",
    "PUSH3",
    "PUSH4",
    "PUSH5",
    "PUSH6",
    "PUSH7",
    "PUSH8",
    "PUSH9",
    "PUSH10",
    "PUSH11",
    "PUSH12",
    "PUSH13",
    "PUSH14",
    "PUSH15",
    "PUSH16",
    "PUSH17",
    "PUSH18",
    "PUSH19",
    "PUSH20",
    "PUSH21",
    "PUSH22",
    "PUSH23",
    "PUSH24",
    "PUSH25",
    "PUSH26",
    "PUSH27",
    "PUSH28",
    "PUSH29",
    "PUSH30",
    "PUSH31",
    "PUSH32",
    "DUP1",
    "DUP2",
    "DUP3",
    "DUP4",
    "DUP5",
    "DUP6",
    "DUP7",
    "DUP8",
    "DUP9",
    "DUP10",
    "DUP11",
    "DUP12",
    "DUP13",
    "DUP14",
    "DUP15",
    "DUP16",
    "SWAP1",
    "SWAP2",
    "SWAP3",
    "SWAP4",
    "SWAP5",
    "SWAP6",
    "SWAP7",
    "SWAP8",
    "SWAP9",
    "SWAP10",
    "SWAP11",
    "SWAP12",
    "SWAP13",
    "SWAP14",
    "SWAP15",
    "SWAP16",
    "LOG0",
    "LOG1",
    "LOG2",
    "LOG3",
    "LOG4",
    "CREATE",
    "CALL",
    "CALLCODE",
    "STATICCALL",
    "RETURN",
    "DELEGATECALL",
    "CREATE2",
    "REVERT",
    "INVALID",
    "SELFDESTRUCT",
]

yul_funcs = [
    "datasize",
    "dataoffset",
    "datacopy",
    "setimmutable",
    "loadimmutable",
]

builtins = [
    x.lower()
    for x in evm_opcodes
    if not (
        x.startswith("PUSH")
        or x.startswith("SWAP")
        or x.startswith("DUP")
        or x == "JUMP"
        or x == "JUMPI"
        or x == "JUMPDEST"
    )
] + yul_funcs

# "identifier": [input_count, output_count]
function_args = {
    "byte": [2, 1],
    "clz": [1, 1],
    "addmod": [3, 1],
    "mulmod": [3, 1],
    "signextend": [2, 1],
    "keccak256": [2, 1],
    "pc": [0, 1],
    "pop": [1, 0],
    "mload": [1, 1],
    "mstore": [2, 0],
    "mstore8": [2, 0],
    "sload": [1, 1],
    "sstore": [2, 0],
    "tload": [1, 1],
    "tstore": [2, 0],
    "msize": [1, 1],
    "gas": [0, 1],
    "address": [0, 1],
    "balance": [1, 1],
    "selfbalance": [0, 1],
    "basefee": [0, 1],
    "blobhash": [1, 1],
    "blobbasefee": [0, 1],
    "caller": [0, 1],
    "callvalue": [0, 1],
    "calldataload": [1, 1],
    "calldatasize": [0, 1],
    "calldatacopy": [3, 0],
    "codesize": [0, 1],
    "codecopy": [3, 0],
    "extcodesize": [1, 1],
    "extcodecopy": [4, 0],
    "returndatasize": [0, 1],
    "returndatacopy": [3, 0],
    "mcopy": [3, 0],
    "extcodehash": [1, 1],
    "create": [3, 1],
    "create2": [4, 1],
    "call": [7, 1],
    "callcode": [7, 1],
    "delegatecall": [6, 1],
    "staticcall": [6, 1],
    "return": [2, 0],
    "revert": [2, 0],
    "selfdestruct": [1, 0],
    "invalid": [0, 0],
    "log0": [2, 0],
    "log1": [3, 0],
    "log2": [4, 0],
    "log3": [5, 0],
    "log4": [6, 0],
    "chainid": [0, 1],
    "origin": [0, 1],
    "gasprice": [0, 1],
    "blockhash": [1, 1],
    "coinbase": [0, 1],
    "timestamp": [0, 1],
    "number": [0, 1],
    "difficulty": [0, 1],
    "prevrandao": [0, 1],
    "gaslimit": [0, 1],
}


def format_function_descriptor(name: str) -> str:
    if name not in function_args:
        return name + "()"

    return name + "(" + ",".join(["uint256"] * function_args[name][0]) + ")"


for k, v in function_args.items():
    SOLIDITY_FUNCTIONS[format_function_descriptor(k)] = ["uint256"] * v[1]

unary_ops = {
    "not": UnaryOperationType.TILD,
    "iszero": UnaryOperationType.BANG,
}

binary_ops = {
    "add": BinaryOperationType.ADDITION,
    "sub": BinaryOperationType.SUBTRACTION,
    "mul": BinaryOperationType.MULTIPLICATION,
    "div": BinaryOperationType.DIVISION,
    "sdiv": BinaryOperationType.DIVISION_SIGNED,
    "mod": BinaryOperationType.MODULO,
    "smod": BinaryOperationType.MODULO_SIGNED,
    "exp": BinaryOperationType.POWER,
    "lt": BinaryOperationType.LESS,
    "gt": BinaryOperationType.GREATER,
    "slt": BinaryOperationType.LESS_SIGNED,
    "sgt": BinaryOperationType.GREATER_SIGNED,
    "eq": BinaryOperationType.EQUAL,
    "and": BinaryOperationType.AND,
    "or": BinaryOperationType.OR,
    "xor": BinaryOperationType.CARET,
    "shl": BinaryOperationType.LEFT_SHIFT,
    "shr": BinaryOperationType.RIGHT_SHIFT,
    "sar": BinaryOperationType.RIGHT_SHIFT_ARITHMETIC,
}


class YulBuiltin:
    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name



"""
Module printing the inheritance relation

The inheritance shows the relation between the contracts
"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.colors import blue, green


class PrinterInheritance(AbstractPrinter):
    ARGUMENT = "inheritance"
    HELP = "Print the inheritance relations between contracts"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#inheritance"

    def _get_child_contracts(self, base):
        # Generate function to get all child contracts of a base contract
        for child in self.contracts:
            if base in child.inheritance:
                yield child

    def output(self, filename):
        """
        Output the inheritance relation

        _filename is not used
        Args:
            _filename(string)
        """
        info = "Inheritance\n"

        info += blue("Child_Contract -> ") + green("Immediate_Base_Contracts")
        info += green(" [Not_Immediate_Base_Contracts]")

        result = {"child_to_base": {}, "paths": {}}

        for child in self.contracts:
            info += blue(f"\n+ {child.name}\n")
            result["paths"][child.name] = child.source_mapping.filename.absolute
            result["child_to_base"][child.name] = {"immediate": [], "not_immediate": []}
            if child.inheritance:
                immediate = child.immediate_inheritance
                not_immediate = [i for i in child.inheritance if i not in immediate]

                info += " -> " + green(", ".join(map(str, immediate))) + "\n"
                result["child_to_base"][child.name]["immediate"] = list(map(str, immediate))
                if not_immediate:
                    info += ", [" + green(", ".join(map(str, not_immediate))) + "]\n"
                    result["child_to_base"][child.name]["not_immediate"] = list(
                        map(str, not_immediate)
                    )

        info += green("\n\nBase_Contract -> ") + blue("Immediate_Child_Contracts") + "\n"
        info += blue(" [Not_Immediate_Child_Contracts]") + "\n"

        result["base_to_child"] = {}
        for base in self.contracts:
            info += green(f"\n+ {base.name}") + "\n"
            children = list(self._get_child_contracts(base))

            result["base_to_child"][base.name] = {"immediate": [], "not_immediate": []}
            if children:
                immediate = [child for child in children if base in child.immediate_inheritance]
                not_immediate = [child for child in children if child not in immediate]

                info += " -> " + blue(", ".join(map(str, immediate))) + "\n"
                result["base_to_child"][base.name]["immediate"] = list(map(str, immediate))
                if not_immediate:
                    info += ", [" + blue(", ".join(map(str, not_immediate))) + "]" + "\n"
                    result["base_to_child"][base.name]["not_immediate"] = list(map(str, immediate))
        self.info(info)

        res = self.generate_output(info, additional_fields=result)

        return res



from typing import Union

from slither.core.declarations import (
    Event,
    Contract,
    SolidityVariableComposed,
    SolidityFunction,
    Structure,
)
from slither.core.declarations.custom_error import CustomError
from slither.core.variables.variable import Variable
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.slithir.operations.member import Member
from slither.slithir.tmp_operations.tmp_new_array import TmpNewArray
from slither.slithir.tmp_operations.tmp_new_contract import TmpNewContract
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.tuple import TupleVariable


class TmpCall(OperationWithLValue):
    def __init__(
        self,
        called: SourceMapping,
        nbr_arguments: int,
        result: TupleVariable | TemporaryVariable,
        type_call: str,
        names: list[str] | None = None,
    ) -> None:
        """
        #### Parameters
        names -
            For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
            Otherwise, None.
        """
        assert isinstance(
            called,
            (
                Contract,
                Variable,
                SolidityVariableComposed,
                SolidityFunction,
                Structure,
                Event,
                CustomError,
            ),
        )
        super().__init__()
        self._called = called
        self._nbr_arguments = nbr_arguments
        self._type_call = type_call
        self._names = names
        self._lvalue = result
        self._ori = None  #
        self._callid = None
        self._gas = None
        self._value = None
        self._salt = None

    @property
    def names(self) -> list[str] | None:
        """
        For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
        Otherwise, None.
        """
        return self._names

    @property
    def call_value(self):
        return self._value

    @call_value.setter
    def call_value(self, v):
        self._value = v

    @property
    def call_gas(self):
        return self._gas

    @call_gas.setter
    def call_gas(self, gas):
        self._gas = gas

    @property
    def call_salt(self):
        return self._salt

    @call_salt.setter
    def call_salt(self, salt):
        self._salt = salt

    @property
    def call_id(self):
        return self._callid

    @call_id.setter
    def call_id(self, c):
        self._callid = c

    @property
    def read(self):
        return [self.called]

    @property
    def called(self):
        return self._called

    @called.setter
    def called(self, c):
        self._called = c

    @property
    def nbr_arguments(self) -> int:
        return self._nbr_arguments

    @property
    def type_call(self) -> str:
        return self._type_call

    @property
    def ori(self) -> Union[TmpNewContract, TmpNewArray, "TmpCall", Member] | None:
        return self._ori

    def set_ori(self, ori: Union["TmpCall", TmpNewContract, TmpNewArray, Member]) -> None:
        self._ori = ori

    def __str__(self):
        return str(self.lvalue) + f" = TMPCALL{self.nbr_arguments} " + str(self._called)



from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.output import Output
from slither.utils.source_mapping import get_definition, get_implementation, get_references


class Declaration(AbstractPrinter):
    ARGUMENT = "declaration"
    HELP = "Prototype showing the source code declaration, implementation and references of the contracts objects"

    WIKI = "TODO"

    def output(self, _filename: str) -> Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""
        for compilation_unit in self.slither.compilation_units:
            txt += "\n# Contracts\n"
            for contract in compilation_unit.contracts:
                txt += f"# {contract.name}\n"
                contract_def = get_definition(contract, compilation_unit.core.crytic_compile)
                txt += f"\t- Declaration: {contract_def.to_detailed_str()}\n"
                txt += f"\t- Implementation(s): {[x.to_detailed_str() for x in list(self.slither.offset_to_implementations(contract.source_mapping.filename.absolute, contract_def.start))]}\n"
                txt += (
                    f"\t- References: {[x.to_detailed_str() for x in get_references(contract)]}\n"
                )

                txt += "\n\t## Function\n"

                for func in contract.functions_declared:
                    txt += f"\t\t- {func.canonical_name}\n"
                    function_def = get_definition(func, compilation_unit.core.crytic_compile)
                    txt += f"\t\t\t- Declaration: {function_def.to_detailed_str()}\n"
                    txt += f"\t\t\t- Implementation(s): {[x.to_detailed_str() for x in list(self.slither.offset_to_implementations(func.source_mapping.filename.absolute, function_def.start))]}\n"
                    txt += f"\t\t\t- References: {[x.to_detailed_str() for x in get_references(func)]}\n"

                txt += "\n\t## State variables\n"

                for var in contract.state_variables:
                    txt += f"\t\t- {var.name}\n"
                    txt += f"\t\t\t- Declaration: {get_definition(var, compilation_unit.core.crytic_compile).to_detailed_str()}\n"
                    txt += f"\t\t\t- Implementation: {get_implementation(var).to_detailed_str()}\n"
                    txt += f"\t\t\t- References: {[x.to_detailed_str() for x in get_references(var)]}\n"

                txt += "\n\t## Structures\n"

                for st in contract.structures:
                    txt += f"\t\t- {st.name}\n"
                    txt += f"\t\t\t- Declaration: {get_definition(st, compilation_unit.core.crytic_compile).txt}\n"
                    txt += f"\t\t\t- Implementation: {get_implementation(st).to_detailed_str()}\n"
                    txt += (
                        f"\t\t\t- References: {[x.to_detailed_str() for x in get_references(st)]}\n"
                    )

        self.info(txt)
        res = self.generate_output(txt)
        return res



from slither.core.cfg.node import Node
from slither.core.declarations import Contract
from slither.core.declarations.function import Function
from slither.core.solidity_types import Type
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import HighLevelCall
from slither.analyses.data_dependency.data_dependency import is_tainted
from slither.utils.output import Output


class ReturnBomb(AbstractDetector):
    ARGUMENT = "return-bomb"
    HELP = "A low level callee may consume all callers gas unexpectedly."
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#return-bomb"

    WIKI_TITLE = "Return Bomb"
    WIKI_DESCRIPTION = "A low level callee may consume all callers gas unexpectedly."
    WIKI_EXPLOIT_SCENARIO = """
```solidity
//Modified from https://github.com/nomad-xyz/ExcessivelySafeCall
contract BadGuy {
    function youveActivateMyTrapCard() external pure returns (bytes memory) {
        assembly{
            revert(0, 1000000)
        }
    }
}

contract Mark {
    function oops(address badGuy) public{
        bool success;
        bytes memory ret;

        // Mark pays a lot of gas for this copy
        //(success, ret) = badGuy.call{gas:10000}(
        (success, ret) = badGuy.call(
            abi.encodeWithSelector(
                BadGuy.youveActivateMyTrapCard.selector
            )
        );

        // Mark may OOG here, preventing local state changes
        //importantCleanup();
    }
}

```
After Mark calls BadGuy bytes are copied from returndata to memory, the memory expansion cost is paid. This means that when using a standard solidity call, the callee can "returnbomb" the caller, imposing an arbitrary gas cost.
Callee unexpectedly makes the caller OOG.
"""

    WIKI_RECOMMENDATION = "Avoid unlimited implicit decoding of returndata."

    @staticmethod
    def is_dynamic_type(ty: Type) -> bool:
        # ty.is_dynamic ?
        name = str(ty)
        if "[]" in name or name in ("bytes", "string"):
            return True
        return False

    def get_nodes_for_function(self, function: Function, contract: Contract) -> list[Node]:
        nodes = []

        for ir in [ir for _, ir in function.high_level_calls] + function.low_level_calls:
            if not is_tainted(ir.destination, contract):  # type:ignore
                # Only interested if the target address is controlled/tainted
                continue

            if isinstance(ir, HighLevelCall) and isinstance(ir.function, Function):
                # in normal highlevel calls return bombs are _possible_
                # if the return type is dynamic and the caller tries to copy and decode large data
                has_dyn = False
                if ir.function.return_type:
                    has_dyn = any(self.is_dynamic_type(ty) for ty in ir.function.return_type)

                if not has_dyn:
                    continue

            # If a gas budget was specified then the
            # user may not know about the return bomb
            if ir.call_gas is None:
                # if a gas budget was NOT specified then the caller
                # may already suspect the call may spend all gas?
                continue

            nodes.append(ir.node)
        # TODO: check that there is some state change after the call

        return nodes

    def _detect(self) -> list[Output]:
        results = []

        for contract in self.compilation_unit.contracts:
            for function in contract.functions_declared:
                nodes = self.get_nodes_for_function(function, contract)
                if nodes:
                    info: DETECTOR_INFO = [
                        function,
                        " tries to limit the gas of an external call that controls implicit decoding\n",
                    ]

                    for node in sorted(nodes, key=lambda x: x.node_id):
                        info += ["\t", node, "\n"]

                    res = self.generate_result(info)
                    results.append(res)

        return results



from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator
from slither.core.expressions.assignment_operation import (
    AssignmentOperationType,
    AssignmentOperation,
)

assignment_operators = [
    AssignmentOperationType.ASSIGN_ADDITION,
    AssignmentOperationType.ASSIGN_SUBTRACTION,
    AssignmentOperationType.ASSIGN,
    AssignmentOperationType.ASSIGN_OR,
    AssignmentOperationType.ASSIGN_CARET,
    AssignmentOperationType.ASSIGN_AND,
    AssignmentOperationType.ASSIGN_LEFT_SHIFT,
    AssignmentOperationType.ASSIGN_RIGHT_SHIFT,
    AssignmentOperationType.ASSIGN_MULTIPLICATION,
    AssignmentOperationType.ASSIGN_DIVISION,
    AssignmentOperationType.ASSIGN_MODULO,
]


class ASOR(AbstractMutator):
    NAME = "ASOR"
    HELP = "Assignment Operator Replacement"

    def _mutate(self) -> dict:
        result: dict = {}

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                for ir in node.irs:
                    if (
                        isinstance(ir.expression, AssignmentOperation)
                        and ir.expression.type in assignment_operators
                    ):
                        if ir.expression.type == AssignmentOperationType.ASSIGN:
                            continue
                        alternative_ops = assignment_operators[:]
                        try:
                            alternative_ops.remove(ir.expression.type)
                        except ValueError:
                            continue
                        for op in alternative_ops:
                            if op != ir.expression:
                                start = node.source_mapping.start
                                stop = start + node.source_mapping.length
                                old_str = node.source_mapping.content
                                line_no = node.source_mapping.lines
                                # Replace the expression with true
                                new_str = f"{old_str.split(str(ir.expression.type))[0]}{op}{old_str.split(str(ir.expression.type))[1]}"
                                create_patch_with_line(
                                    result,
                                    self.in_file,
                                    start,
                                    stop,
                                    old_str,
                                    new_str,
                                    line_no[0],
                                )
        return result
